from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pendulum
from pyspark.sql.types import StructType
from typeguard import typechecked

from tecton import DynamoConfig
from tecton import RedisConfig
from tecton._internals.fco import Fco
from tecton._internals.feature_definition import FeatureDefinition
from tecton.basic_info import prepare_basic_info
from tecton.entities.entity import Entity
from tecton.entities.entity import OverriddenEntity
from tecton.features_common.feature_configs import DatabricksClusterConfig
from tecton.features_common.feature_configs import DeltaConfig
from tecton.features_common.feature_configs import EMRClusterConfig
from tecton.features_common.feature_configs import ExistingClusterConfig
from tecton_proto.args.feature_view_pb2 import EntityKeyOverride
from tecton_proto.args.feature_view_pb2 import FeatureTableArgs
from tecton_proto.args.feature_view_pb2 import FeatureViewArgs
from tecton_proto.args.feature_view_pb2 import FeatureViewType
from tecton_spark.id_helper import IdHelper
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper
from tecton_spark.time_utils import strict_pytimeparse


class FeatureTable(FeatureDefinition):
    """
    Declare a FeatureTable.

    The FeatureTable class is used to represent one or many features that are pushed to Tecton from external feature computation systems.
    """

    @typechecked
    def __init__(
        self,
        *,
        name: str,
        entities: List[Union[Entity, OverriddenEntity]],
        schema: StructType,
        ttl: str,
        online: Optional[bool] = False,
        offline: Optional[bool] = False,
        description: Optional[str] = None,
        owner: Optional[str] = None,
        family: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        offline_config: DeltaConfig = DeltaConfig(),
        online_config: Optional[Union[DynamoConfig, RedisConfig]] = None,
        batch_cluster_config: Optional[Union[ExistingClusterConfig, DatabricksClusterConfig, EMRClusterConfig]] = None,
        online_serving_index: Optional[List[str]] = None,
    ):
        """
        Instantiates a new FeatureTable.

        :param name: Unique, human friendly name that identifies the FeatureTable.
        :param entities: A list of Entity objects, used to organize features.
        :param schema: A Spark schema definition (StructType) for the FeatureTable.
            Supported types are: LongType, DoubleType, StringType, BooleanType and TimestampType (for inferred timestamp column only).
        :param ttl: The TTL (or "look back window") for features defined by this feature table. This parameter determines how long features will live in the online store and how far to  "look back" relative to a training example's timestamp when generating offline training sets. Shorter TTLs improve performance and reduce costs.
        :param online: Enable writing to online feature store. (Default: False)
        :param offline: Enable writing to offline feature store. (Default: False)
        :param description: Human readable description.
        :param owner: Owner name (typically the email of the primary maintainer).
        :param family: Family of this Feature Table, used to group Tecton Objects.
        :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
        :param offline_config: Configuration for how data is written to the offline feature store.
        :param online_config: Configuration for how data is written to the online feature store.
        :param batch_cluster_config: Batch materialization cluster configuration. Should be one of:
            [``EMRClusterConfig``, ``DatabricksClusterConfig``, ``ExistingClusterConfig``]
        :param online_serving_index: (Advanced) Defines the set of join keys that will be indexed and queryable during online serving.
            Defaults to the complete set of join keys. Up to one join key may be omitted. If one key is omitted, online requests to a Feature Service will
            return all feature vectors that match the specified join keys.
        :returns: A Feature Table

        An example declaration of a FeatureTable

        .. code-block:: python

            from pyspark.sql.types import StructType, StructField, LongType, StringType, TimestampType
            from tecton import Entity, FeatureTable

            # Declare your user Entity instance here or import it if defined elsewhere in
            # your Tecton repo.
            user = ...

            # Schema for your feature table
            schema = StructType([
                StructField('user_id', StringType()),
                StructField('timestamp', TimestampType()),
                StructField('user_login_count_7d', LongType()),
                StructField('user_login_count_30d', LongType())
            ])

            user_login_counts = FeatureTable(
                name='user_login_counts',
                entities=[user],
                schema=schema,
                online=True,
                offline=True,
                ttl='30day'
            )
        """
        from tecton.cli.common import get_fco_source_info

        self._source_info = get_fco_source_info()
        basic_info = prepare_basic_info(name=name, description=description, owner=owner, family=family, tags=tags)

        self._args = FeatureViewArgs()

        self._args.feature_view_id.CopyFrom(IdHelper.from_string(IdHelper.generate_string_id()))
        self._args.info.CopyFrom(basic_info)
        self._args.feature_view_type = FeatureViewType.FEATURE_VIEW_TYPE_FEATURE_TABLE

        self._args.entities.extend(
            [EntityKeyOverride(entity_id=entity._id(), join_keys=entity.join_keys) for entity in entities]
        )
        if online_serving_index:
            self._args.online_serving_index.extend(online_serving_index)

        self._args.online_enabled = online
        self._args.offline_enabled = offline

        feature_table_args = FeatureTableArgs()
        feature_table_args.output_schema.CopyFrom(SparkSchemaWrapper(schema).to_proto())
        if ttl:
            feature_table_args.serving_ttl.FromTimedelta(pendulum.duration(seconds=strict_pytimeparse(ttl)))
        if batch_cluster_config:
            cluster_config = batch_cluster_config._to_cluster_proto()
            feature_table_args.batch_materialization.CopyFrom(cluster_config)
        feature_table_args.offline_config.CopyFrom(offline_config._to_proto())
        if online_config:
            feature_table_args.online_store_config.CopyFrom(online_config._to_proto())
        self._args.feature_table_args.CopyFrom(feature_table_args)

        Fco._register(self)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/column_type.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/common/column_type.proto',
  package='tecton_proto.common',
  syntax='proto2',
  serialized_options=b'\n\021com.tecton.commonP\001Z(github.com/tecton-ai/tecton_proto/common',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n%tecton_proto/common/column_type.proto\x12\x13tecton_proto.common*\x84\x02\n\nColumnType\x12\x17\n\x13\x43OLUMN_TYPE_UNKNOWN\x10\x00\x12\x15\n\x11\x43OLUMN_TYPE_INT64\x10\x01\x12\x16\n\x12\x43OLUMN_TYPE_DOUBLE\x10\x02\x12\x16\n\x12\x43OLUMN_TYPE_STRING\x10\x03\x12\x14\n\x10\x43OLUMN_TYPE_BOOL\x10\x06\x12\x1c\n\x18\x43OLUMN_TYPE_STRING_ARRAY\x10\x07\x12\x1b\n\x17\x43OLUMN_TYPE_INT64_ARRAY\x10\x08\x12\x1c\n\x18\x43OLUMN_TYPE_DOUBLE_ARRAY\x10\t\x12\x1b\n\x17\x43OLUMN_TYPE_FLOAT_ARRAY\x10\n\"\x04\x08\x04\x10\x04\"\x04\x08\x05\x10\x05\x42?\n\x11\x63om.tecton.commonP\x01Z(github.com/tecton-ai/tecton_proto/common'
)

_COLUMNTYPE = _descriptor.EnumDescriptor(
  name='ColumnType',
  full_name='tecton_proto.common.ColumnType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='COLUMN_TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN_TYPE_INT64', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN_TYPE_DOUBLE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN_TYPE_STRING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN_TYPE_BOOL', index=4, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN_TYPE_STRING_ARRAY', index=5, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN_TYPE_INT64_ARRAY', index=6, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN_TYPE_DOUBLE_ARRAY', index=7, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COLUMN_TYPE_FLOAT_ARRAY', index=8, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=63,
  serialized_end=323,
)
_sym_db.RegisterEnumDescriptor(_COLUMNTYPE)

ColumnType = enum_type_wrapper.EnumTypeWrapper(_COLUMNTYPE)
COLUMN_TYPE_UNKNOWN = 0
COLUMN_TYPE_INT64 = 1
COLUMN_TYPE_DOUBLE = 2
COLUMN_TYPE_STRING = 3
COLUMN_TYPE_BOOL = 6
COLUMN_TYPE_STRING_ARRAY = 7
COLUMN_TYPE_INT64_ARRAY = 8
COLUMN_TYPE_DOUBLE_ARRAY = 9
COLUMN_TYPE_FLOAT_ARRAY = 10


DESCRIPTOR.enum_types_by_name['ColumnType'] = _COLUMNTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from typing import List
from typing import Optional
from typing import Tuple

import attr

from tecton_proto.data.feature_service_pb2 import FeatureSetItem
from tecton_spark.fco_container import FcoContainer
from tecton_spark.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_spark.id_helper import IdHelper


@attr.s(frozen=True, auto_attribs=True)
class _Key(object):
    name: str
    namespace: str


@attr.s(frozen=True, auto_attribs=True)
class FeatureDefinitionAndJoinConfig(object):
    """
    A feature definition and its associated join configuration.

    :param feature_definition: FeatureView/FeatureTable object.
    :param join_keys: The mapping from FeatureService's join keys to FeatureView/FeatureTable's join keys.
    :param namespace: The namespace.
    """

    feature_definition: FeatureDefinition
    name: str
    join_keys: List[Tuple[str, str]]
    namespace: str
    features: List[str]

    @classmethod
    def _from_proto(cls, feature_set_item: FeatureSetItem, fco_container: FcoContainer):
        """
        :param feature_set_item: FSI proto
        :param fco_container: Contains all FSI dependencies (transitively), e.g., FV, Entities, DS-es, Transformations
        """

        join_keys = [(i.spine_column_name, i.package_column_name) for i in feature_set_item.join_configuration_items]
        fv_proto = fco_container.get_by_id(IdHelper.to_string(feature_set_item.feature_view_id))
        feature_definition = FeatureDefinition(fv_proto, fco_container)
        return FeatureDefinitionAndJoinConfig(
            feature_definition=feature_definition,
            name=feature_definition.name,
            join_keys=join_keys,
            namespace=feature_set_item.namespace,
            features=list(feature_set_item.feature_columns),
        )

    def _key(self) -> _Key:
        return _Key(namespace=self.namespace or "", name=self.name)


class FeatureSetConfig(object):
    """
    Config used to create a :class:`FeatureService`.
    """

    # This stores all the features + all the odfv dependent features
    _definitions_and_configs: List[FeatureDefinitionAndJoinConfig]

    def __init__(self):
        """
        Initialize a new FeatureSetConfig.
        """

        self._definitions_and_configs = []

    def _add(
        self,
        fd: FeatureDefinition,
        namespace: Optional[str] = None,
    ):
        if namespace is None:
            namespace = fd.name

        fd_and_config = FeatureSetConfig._make_definition_and_config(fd, namespace=namespace)
        self._definitions_and_configs.append(fd_and_config)

    @property  # type: ignore
    def feature_definitions(self) -> List[FeatureDefinition]:
        """
        Returns the FeatureViews/FeatureTables enclosed in this FeatureSetConfig.
        """
        return [config.feature_definition for config in self._get_feature_definitions_and_join_configs()]

    @property  # type: ignore
    def features(self) -> List[str]:
        """
        Returns the features generated by the enclosed feature definitions.
        """
        return [
            features
            for config in self._get_feature_definitions_and_join_configs()
            for features in FeatureSetConfig._get_full_feature_names(config)
        ]

    @staticmethod
    def _get_full_feature_names(config: FeatureDefinitionAndJoinConfig):
        return [config.namespace + "." + feature if config.namespace else feature for feature in config.features]

    @staticmethod
    def _make_definition_and_config(fd: FeatureDefinition, namespace: str) -> FeatureDefinitionAndJoinConfig:
        join_keys = [(join_key, join_key) for join_key in fd.join_keys]
        features = fd.features

        return FeatureDefinitionAndJoinConfig(
            feature_definition=fd, name=fd.name, join_keys=join_keys, namespace=namespace, features=features
        )

    def _get_feature_definitions_and_join_configs(self) -> List[FeatureDefinitionAndJoinConfig]:
        """
        Returns a list of all feature definitions and their configs in the insertion order.
        """
        return self._definitions_and_configs

    @classmethod
    def _from_protos(cls, feature_set_items, fco_container: FcoContainer):
        """
        :param feature_set_items: FSI protos
        :param fco_container: Contains all FSI dependencies (transitively), e.g., FVs, Entities, DS-es, Transformations
        """
        config = FeatureSetConfig()
        config._definitions_and_configs = [
            FeatureDefinitionAndJoinConfig._from_proto(p, fco_container) for p in feature_set_items
        ]
        return config



from dataclasses import dataclass
from textwrap import dedent
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import pandas

from tecton_proto.args.new_transformation_pb2 import NewTransformationArgs as TransformationArgs
from tecton_proto.args.new_transformation_pb2 import TransformationMode
from tecton_proto.args.pipeline_pb2 import DataSourceNode
from tecton_proto.args.pipeline_pb2 import Pipeline
from tecton_proto.args.pipeline_pb2 import PipelineNode
from tecton_proto.args.pipeline_pb2 import TransformationNode
from tecton_proto.data.new_transformation_pb2 import NewTransformation as Transformation
from tecton_proto.data.virtual_data_source_pb2 import VirtualDataSource
from tecton_spark import function_serialization
from tecton_spark.id_helper import IdHelper
from tecton_spark.pipeline_common import constant_node_to_value
from tecton_spark.pipeline_common import CONSTANT_TYPE
from tecton_spark.pipeline_common import get_keyword_inputs
from tecton_spark.pipeline_common import get_transformation_name
from tecton_spark.pipeline_common import positional_inputs
from tecton_spark.pipeline_common import transformation_type_checker
from tecton_spark.snowflake.errors import TectonSnowflakeNotImplementedError
from tecton_spark.snowflake.templates_utils import load_template

PIPELINE_TEMPLATE = None
TEMP_CTE_PREFIX = "_TT_CTE_"
SPINE_TABLE_NAME = "_TT_SPINE_TABLE"
UDF_OUTPUT_COLUMN_NAME = "_TT_UDF_OUTPUT"

# TODO(TEC-6893): We should use Snowpark struct types in the repo directly.
SPARK_TO_NATIVE_TYPES = {
    "long": "int",
    "double": "float",
    "string": "str",
    "boolean": "bool",
    "long_array": "list",
    "float_array": "list",
    "double_array": "list",
    "string_array": "list",
}

SPARK_TO_SNOWFLAKE_TYPES = {
    "long": "NUMBER",
    "double": "FLOAT",
    "string": "STRING",
    "boolean": "BOOLEAN",
    "long_array": "ARRAY",
    "float_array": "ARRAY",
    "double_array": "ARRAY",
    "string_array": "ARRAY",
}


def _load_template():
    # TODO: Do this at module loading time once we sort out including the templates in the public SDK build
    global PIPELINE_TEMPLATE
    if not PIPELINE_TEMPLATE:
        PIPELINE_TEMPLATE = load_template("transformation_pipeline.sql.j2")


def pipeline_to_sql_string(
    pipeline: Pipeline,
    data_sources: List[VirtualDataSource],
    transformations: List[Transformation],
) -> str:
    _load_template()
    return _PipelineBuilder(
        pipeline,
        data_sources,
        transformations,
    ).get_sql_string()


# Pandas Pipeline (ODFV)
# input_df (snowpark df) is the spine passed in by the user (including request context),
# and it has been augmented with dependent fv fields in of the form "_udf_internal_{input_name}.{feature_field_name}".
# The dataframe we return will be everything from the spine, with the on-demand features added
# TODO: Figure out a way to get the dependent fv fields on tecton apply, this way we can
#      register the udf on apply and avoid doing it adhoc on feature retrival.
def pipeline_to_df_with_input(
    session: "snowflake.snowpark.Session",
    # This should have data from all inputs
    input_df: "snowflake.snowpark.DataFrame",
    pipeline: Pipeline,
    transformations: List[Transformation],
    output_schema: Dict[str, str],
    name: str,
    fv_id: str,
) -> "snowflake.snowpark.DataFrame":
    _load_template()
    # TODO: Currently there's a bug in toPandas() call, types may not be casted to the correct type.
    # e.g. Long is currently being casted to object(decimal.Decimal) instead of int64.
    return _ODFVPipelineBuilder(
        session=session,
        input_df=input_df,
        output_schema=output_schema,
        name=name,
        pipeline=pipeline,
        transformations=transformations,
        fv_id=fv_id,
    ).get_df()


# Used for j2 template
@dataclass
class _NodeInput:
    name: str
    sql_str: str


# This class is for Snowflake pipelines
class _PipelineBuilder:
    # The value of internal nodes in the tree
    _VALUE_TYPE = Union[str, CONSTANT_TYPE]

    def __init__(
        self,
        pipeline: Pipeline,
        data_sources: List[VirtualDataSource],
        # we only use mode and name from these
        transformations: Union[List[Transformation], List[TransformationArgs]],
    ):
        self._pipeline = pipeline
        self._id_to_ds = {IdHelper.to_string(ds.virtual_data_source_id): ds for ds in data_sources}
        self._id_to_transformation = {IdHelper.to_string(t.transformation_id): t for t in transformations}

    def get_sql_string(self) -> str:
        sql_str = self._node_to_value(self._pipeline.root)
        assert isinstance(sql_str, str)
        return sql_str

    def _node_to_value(self, pipeline_node: PipelineNode) -> _VALUE_TYPE:
        if pipeline_node.HasField("transformation_node"):
            return self._transformation_node_to_sql_str(pipeline_node.transformation_node)
        elif pipeline_node.HasField("data_source_node"):
            return self._data_source_node_to_sql_str(pipeline_node.data_source_node)
        elif pipeline_node.HasField("constant_node"):
            return constant_node_to_value(pipeline_node.constant_node)
        elif pipeline_node.HasField("materialization_context_node"):
            raise TectonSnowflakeNotImplementedError(
                "MaterializationContext is not supported in Snowflake SQL pipelines"
            )
        elif pipeline_node.HasField("request_data_source_node"):
            raise TectonSnowflakeNotImplementedError("RequestDataSource is not supported in Snowflake SQL pipelines")
        elif pipeline_node.HasField("feature_view_node"):
            raise TectonSnowflakeNotImplementedError(
                "Dependent FeatureViews are not supported in Snowflake SQL pipelines"
            )
        else:
            raise KeyError(f"Unknown PipelineNode type: {pipeline_node}")

    def _data_source_node_to_sql_str(self, data_source_node: DataSourceNode) -> str:
        """Creates a sql string from a ds and time parameters."""
        ds = self._id_to_ds[IdHelper.to_string(data_source_node.virtual_data_source_id)]
        # TODO(TEC-7274): Applies time window on the data source
        return self._get_ds_sql_str(ds)

    def _get_ds_sql_str(self, ds: VirtualDataSource) -> str:
        # TODO(TEC-5907): Supports other types of data source
        if ds.HasField("batch_data_source"):
            batch_data_source = ds.batch_data_source
            # TODO(TEC-5907): Supports other types of batch data source
            if batch_data_source.HasField("snowflake"):
                snowflake_args = batch_data_source.snowflake.snowflakeArgs
                if snowflake_args.HasField("table"):
                    # Makes sure we have all the info for the table
                    assert snowflake_args.HasField("database")
                    assert snowflake_args.HasField("schema")
                    sql_str = f"{snowflake_args.database}.{snowflake_args.schema}.{snowflake_args.table}"
                else:
                    raise TectonSnowflakeNotImplementedError(
                        f"Snowflake SQL pipeline does not support query as a batch data source"
                    )
            else:
                raise TectonSnowflakeNotImplementedError(
                    f"Snowflake SQL pipeline does not support batch data source: {ds.batch_data_source}"
                )
        else:
            raise TectonSnowflakeNotImplementedError("Snowflake SQL pipeline only supports batch data source")
        return sql_str

    def _transformation_node_to_sql_str(self, transformation_node: TransformationNode) -> str:
        """Recursively translates inputs to values and then passes them to the transformation."""
        args: List[_VALUE_TYPE] = []
        kwargs = {}
        for transformation_input in transformation_node.inputs:
            node_value = self._node_to_value(transformation_input.node)
            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                raise KeyError(f"Unknown argument type for Input node: {transformation_input}")

        return self._apply_transformation_function(transformation_node, args, kwargs)

    def _apply_transformation_function(self, transformation_node, args, kwargs) -> str:
        """For the given transformation node, returns the corresponding sql string."""
        transformation = self._id_to_transformation[IdHelper.to_string(transformation_node.transformation_id)]
        user_function = function_serialization.from_proto(transformation.user_function)
        transformation_name = get_transformation_name(transformation)

        if transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_SNOWFLAKE_SQL:
            return self._wrap_sql_function(transformation_node, user_function)(*args, **kwargs)
        else:
            raise KeyError(
                f"Invalid transformation mode: {TransformationMode.Name(transformation.transformation_mode)} for a Snowflake SQL pipeline"
            )

    def _wrap_sql_function(
        self, transformation_node: TransformationNode, user_function: Callable[..., str]
    ) -> Callable[..., str]:
        def wrapped(*args, **kwargs):
            transformationInputs = []
            wrapped_args = []
            for i, (arg, node_input) in enumerate(zip(args, positional_inputs(transformation_node))):
                input_str, is_sql = self._wrap_node_inputvalue(node_input, arg)
                cte_name = TEMP_CTE_PREFIX + str(i)
                if is_sql:
                    wrapped_args.append(cte_name)
                    transformationInputs.append(_NodeInput(name=cte_name, sql_str=input_str))
                else:
                    wrapped_args.append(input_str)
            keyword_inputs = get_keyword_inputs(transformation_node)
            wrapped_kwargs = {}
            for k, v in kwargs.items():
                node_input = keyword_inputs[k]
                input_str, is_sql = self._wrap_node_inputvalue(node_input, v)
                if is_sql:
                    cte_name = TEMP_CTE_PREFIX + k
                    wrapped_kwargs[k] = cte_name
                    transformationInputs.append(_NodeInput(name=cte_name, sql_str=input_str))
                else:
                    wrapped_kwargs[k] = input_str
            user_function_sql = dedent(user_function(*wrapped_args, **wrapped_kwargs))
            sql_str = PIPELINE_TEMPLATE.render(inputs=transformationInputs, user_function=user_function_sql)
            transformation_name = get_transformation_name(
                self._id_to_transformation[IdHelper.to_string(transformation_node.transformation_id)]
            )
            transformation_type_checker(transformation_name, sql_str, "snowflake_sql", self._possible_modes())
            return sql_str

        return wrapped

    def _wrap_node_inputvalue(self, node_input, value: _VALUE_TYPE) -> Tuple[Union[CONSTANT_TYPE], bool]:
        """Returns the node value, along with a boolean indicating whether the input is a sql str."""
        if node_input.node.HasField("constant_node"):
            assert (
                isinstance(value, str)
                or isinstance(value, int)
                or isinstance(value, float)
                or isinstance(value, bool)
                or value is None
            )
            return value, False
        elif node_input.node.HasField("data_source_node"):
            # For data source we don't want a bracket around it
            assert isinstance(value, str)
            return value, False
        else:
            # This should be a sql string already, we need to return this with a bracket wrapping it
            # The current implementation will add a round bracket () to all subquery
            assert isinstance(value, str)
            return f"({value})", True

    def _possible_modes(self):
        return ["snowflake_sql", "pipeline"]


# This class is for Pandas pipelines
class _ODFVPipelineBuilder:
    def __init__(
        self,
        session: "snowflake.snowpark.Session",
        name: str,
        transformations: List[Transformation],
        output_schema: Dict[str, str],
        pipeline: Pipeline,
        fv_id: str,
        input_df: "snowflake.snowpark.DataFrame" = None,
    ):
        self._input_df = input_df
        self._session = session
        self._pipeline = pipeline
        self._name = name
        self._fv_id = fv_id
        self._id_to_transformation = {IdHelper.to_string(t.transformation_id): t for t in transformations}
        self._output_schema = output_schema

    def get_df(self) -> "snowflake.snowpark.DataFrame":
        if not self._pipeline.root.HasField("transformation_node"):
            raise ValueError("Root pipeline has to be a transformation for pandas mode")
        return self._transformation_node_to_df(self._pipeline.root.transformation_node)

    def _transformation_node_to_df(self, transformation_node: TransformationNode) -> "snowflake.snowpark.DataFrame":
        # Columns in snowflake dataframe have double quotes around them.
        udf_args = [c.strip('"') for c in self._input_df.columns if ("_UDF_INTERNAL" in c)]
        input_columns = []
        input_map = {}
        prefix_map = {}
        # Input for On-Demand can only be a feature view, or request data source
        for transformation_input in transformation_node.inputs:
            input_node = transformation_input.node
            if input_node.HasField("feature_view_node"):
                features = []
                feature_view_node = input_node.feature_view_node
                prefix = f"_UDF_INTERNAL_{feature_view_node.input_name}_{self._fv_id}__".upper()
                for feature in udf_args:
                    if not feature.startswith(prefix):
                        continue
                    input_columns.append(feature)
                    features.append(feature)
                input_map[feature_view_node.input_name] = features
                prefix_map[feature_view_node.input_name] = prefix
            elif input_node.HasField("request_data_source_node"):
                request_data_source_node = input_node.request_data_source_node
                field_names = [field.name for field in request_data_source_node.request_context.schema.fields]
                for input_col in field_names:
                    input_columns.append(input_col)
                input_map[request_data_source_node.input_name] = field_names
                prefix_map[request_data_source_node.input_name] = ""
                # Request context should be in the input_df already
            else:
                raise TectonSnowflakeNotImplementedError(
                    "Snowflake only supports feature view and request data source as input."
                )
        # Get back the name of the UDF
        ondemand_udf = self._generate_on_demand_udf(transformation_node, input_map, prefix_map, input_columns)
        return self._call_udf(ondemand_udf, input_columns, self._input_df)

    def _call_udf(
        self,
        ondemand_udf: "snowflake.snowpark.udf.UserDefinedFunction",
        input_columns: List[str],
        input_df: "snowflake.snowpark.DataFrame",
    ) -> "snowflake.snowpark.DataFrame":
        from snowflake.snowpark.functions import array_construct, col

        output_df = input_df.withColumn(UDF_OUTPUT_COLUMN_NAME, ondemand_udf(array_construct(*input_columns)))
        for column in self._output_schema.keys():
            output_df = output_df.withColumn(
                column, col(UDF_OUTPUT_COLUMN_NAME)[column].cast(SPARK_TO_SNOWFLAKE_TYPES[self._output_schema[column]])
            )
        return output_df.select(*input_df.columns, *self._output_schema.keys())

    def _generate_on_demand_udf(
        self,
        transformation_node: TransformationNode,
        input_map: Dict[str, List[str]],
        prefix_map: Dict[str, str],
        input_columns: List[str],
    ) -> "snowflake.snowpark.udf.UserDefinedFunction":
        """Returns the name of the registered udf"""
        transformation = self._id_to_transformation[IdHelper.to_string(transformation_node.transformation_id)]
        serialized_user_fn = transformation.user_function
        # TODO(TEC-6901): This line is added to fix a bug with spark udf. We don't need this in Snowflake.
        serialized_user_fn.body = serialized_user_fn.body.replace(
            "from tecton.interactive.data_frame import DataFrame", ""
        )
        user_function = function_serialization.from_proto(serialized_user_fn)

        use_pandas = transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PANDAS

        def _consume_prefix(s: str, prefix: str) -> str:
            assert s.startswith(prefix), f"{prefix} is not a prefix of {s}"
            return s[len(prefix) :]

        def udf(params: List) -> Dict:
            inputs = {}
            if use_pandas:
                all_inputs_df = pandas.DataFrame([params], columns=input_columns)
                for input_name, columns in input_map.items():
                    df = all_inputs_df[columns]
                    df.columns = df.columns.str[len(prefix_map[input_name]) :]
                    inputs[input_name] = df
            else:
                all_inputs = dict(zip(input_columns, params))
                for input_name, columns in input_map.items():
                    inputs[input_name] = {
                        _consume_prefix(column, prefix_map[input_name]): all_inputs[column] for column in columns
                    }

            result = user_function(**inputs)

            if use_pandas:
                result_df = result.astype("object")
                return {col: result_df[col][0] for col in result_df.columns}
            return result

        function_name = f"_{self._name}"
        self._session.add_packages("pandas")
        return self._session.udf.register(udf, name=function_name, replace=True)



# nopycln: file
from tecton_spark import conf as _conf

if not _conf.get_or_none("ALPHA_SNOWFLAKE_COMPUTE_ENABLED"):
    from tecton.feature_views.feature_view import batch_feature_view
    from tecton.feature_views.feature_view import batch_window_aggregate_feature_view
from tecton.feature_views.feature_view import FeatureDefinition
from tecton.feature_views.feature_view import Input
from tecton.feature_views.feature_view import MaterializedFeatureView
from tecton.feature_views.feature_view import on_demand_feature_view
from tecton.feature_views.feature_view import OnDemandFeatureView
from tecton.feature_views.feature_view import stream_feature_view
from tecton.feature_views.feature_view import stream_window_aggregate_feature_view
from tecton_spark.pipeline_helper import RequestDataSource



import atexit
import logging
import os
import traceback
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import Optional

import pendulum

from tecton import conf
from tecton import version
from tecton._internals import metadata_service
from tecton._internals.env_utils import get_current_username
from tecton_proto.amplitude import client_logging_pb2
from tecton_proto.metadataservice.metadata_service_pb2 import IngestAnalyticsRequest
from tecton_proto.metadataservice.metadata_service_pb2 import IngestClientLogsRequest
from tecton_spark.logger import get_logger
from tecton_spark.logger import get_logging_level


SINGLE_POSITIONAL_RETURN_VALUE_NAME = "0"

logger = get_logger("AnalyticsLogger")


@dataclass
class TectonStateUpdateMetrics:
    """Metrics to log to Amplitude for tecton plan, apply, upgrade, delete."""

    num_total_fcos: int
    num_fcos_changed: int
    error_message: Optional[str] = None

    @classmethod
    def from_error_message(cls, error_message: str):
        return cls(-1, -1, error_message)


class AnalyticsLogger(object):
    """
    Conventions for logging parameters and return values:
    - `self` object of instance methods is logged under "self" name
    - Positional return values are logged under position index names, e.g., "0", "1", etc.
    """

    _shutting_down = False

    def __init__(self):
        # This executor submits requests to MDS via RPC in a different thread; however when the process is terminated
        # the executor joins on all outstanding threads and this may but the process in a bad state and
        # prevent shutdown. This happens most commonly on CI environments, where we don't need analytics events.
        # So we choose to disable the analytics logging there.
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._enabled = os.getenv("TEST_WORKSPACE") is None

        def cleanup():
            self._shutting_down = True

        atexit.register(cleanup)

    def _submit(self, request: IngestClientLogsRequest):
        if not self._enabled:
            return
        self._executor.submit(self._send_sdk_logs_to_mds, request)

    def log_method_entry(self, trace_id, obj, method, args, kwargs, arg_names):
        try:
            request = self._create_method_entry_request(trace_id, obj, method, args, kwargs, arg_names)
            self._submit(request)
        except Exception as e:
            logger.debug("Logging error: %s", e)

    def log_method_return(self, trace_id, obj, method, return_value, execution_time, exception: Optional[Exception]):
        try:
            request = self._create_method_return_request(trace_id, obj, method, return_value, execution_time, exception)
            self._submit(request)
        except Exception as e:
            logger.debug("Logging error: %s", e)

    @staticmethod
    def _create_base_request(trace_id, obj, method) -> IngestClientLogsRequest:
        request = IngestClientLogsRequest()
        data = request.sdk_method_invocation
        data.workspace = conf.get_or_none("TECTON_WORKSPACE")
        data.trace_id = trace_id
        data.time.FromDatetime(pendulum.now("UTC"))
        data.user_id = get_current_username()
        data.log_level = logging.getLevelName(get_logging_level())
        data.sdk_version = version.get_version() or ""
        if obj:
            method_class = obj if isinstance(obj, type) else type(obj)
            data.class_name = method_class.__name__
            data.method_name = method.__name__
        else:
            data.method_name = method.__qualname__
        return request

    @staticmethod
    def _create_method_entry_request(trace_id, obj, method, args, kwargs, arg_names) -> IngestClientLogsRequest:
        request = AnalyticsLogger._create_base_request(trace_id, obj, method)
        data = request.sdk_method_invocation
        data.type = client_logging_pb2.CLIENT_LOG_MESSAGE_TYPE_METHOD_ENTRY
        is_class_method = isinstance(obj, type)
        if not is_class_method:
            # the method might be a constructor, in that case __str__ may not yet work in an
            # uninitialized object
            try:
                self_str = str(obj)
            except Exception as e:
                self_str = f"<failed to convert {type(obj)} to str>"
            AnalyticsLogger._add_value(data, "self", self_str)

        AnalyticsLogger._add_values(data, args, arg_names)
        AnalyticsLogger._add_values(data, kwargs)
        return request

    @staticmethod
    def _create_method_return_request(
        trace_id, obj, method, return_value, execution_time, exception: Optional[Exception]
    ) -> IngestClientLogsRequest:
        request = AnalyticsLogger._create_base_request(trace_id, obj, method)
        data = request.sdk_method_invocation
        data.type = client_logging_pb2.CLIENT_LOG_MESSAGE_TYPE_METHOD_RETURN
        data.execution_time.FromTimedelta(execution_time)
        AnalyticsLogger._add_values(data, return_value)
        if exception:
            AnalyticsLogger._set_error(data.error, exception)
        return request

    def _send_sdk_logs_to_mds(self, request: IngestClientLogsRequest):
        if self._shutting_down:
            return
        try:
            metadata_service.instance().IngestClientLogs(request, timeout=5.0)
        except Exception as e:
            logger.debug("Logging error: %s", e)

    @staticmethod
    def _add_values(data: client_logging_pb2.SDKMethodInvocation, values, positional_value_names=None):
        if values is None:
            return
        if isinstance(values, str):
            AnalyticsLogger._add_value(data, SINGLE_POSITIONAL_RETURN_VALUE_NAME, values)
        elif isinstance(values, dict):
            for k, v in values.items():
                AnalyticsLogger._add_value(data, k, v)
        elif isinstance(values, Iterable):
            for i, value in enumerate(values):
                name = positional_value_names[i] if positional_value_names else str(i)
                AnalyticsLogger._add_value(data, name, value)
        else:
            AnalyticsLogger._add_value(data, SINGLE_POSITIONAL_RETURN_VALUE_NAME, values)

    @staticmethod
    def _add_value(data: client_logging_pb2.SDKMethodInvocation, name, value):
        value_record = data.params_or_return_values.add()
        value_record.name = name
        value_record.value = str(value)

    @staticmethod
    def _set_error(data: client_logging_pb2.ErrorLog, e: BaseException):
        data.message = str(e)
        data.stacktrace = str(traceback.format_tb(e.__traceback__))
        if e.__cause__:
            AnalyticsLogger._set_error(data.cause, e.__cause__)

    def log_cli_event(
        self,
        event_name: str,
        execution_time: pendulum.Period,
        state_update_event: Optional[TectonStateUpdateMetrics] = None,
    ):
        request = IngestAnalyticsRequest()
        request.workspace = conf.get_or_none("TECTON_WORKSPACE")
        event = request.events.add()
        # We do not populate user_id since it is not available in the CLI, we will backfill it in the backend
        event.event_type = event_name
        event.platform = "CLI"
        event.timestamp = int(pendulum.now("UTC").timestamp()) * 1000
        event.event_properties.sdk_version = version.get_version() or ""
        event.event_properties.execution_time.FromTimedelta(execution_time)
        if state_update_event:
            if state_update_event.error_message:
                event.event_properties.success = False
                event.event_properties.error_message = state_update_event.error_message
            else:
                event.event_properties.success = True
                event.event_properties.num_total_fcos = state_update_event.num_total_fcos
                event.event_properties.num_fcos_changed = state_update_event.num_fcos_changed
        # Not using the async threadpool since CLI is a not a long-running process and we would lose events on exit
        self._send_cli_logs_to_mds(request)

    def _send_cli_logs_to_mds(self, request: IngestAnalyticsRequest):
        if self._shutting_down:
            return
        try:
            metadata_service.instance().IngestAnalytics(request, timeout=5.0)
        except Exception as e:
            logger.debug("Logging error (cli): %s", e)



import sys

import click

import tecton
from tecton._internals import metadata_service
from tecton.cli import printer
from tecton.cli.command import TectonGroup
from tecton_proto.metadataservice.metadata_service_pb2 import CreateApiKeyRequest
from tecton_proto.metadataservice.metadata_service_pb2 import DeleteApiKeyRequest
from tecton_proto.metadataservice.metadata_service_pb2 import ListApiKeysRequest
from tecton_spark.id_helper import IdHelper


# TODO(fix the "readonly" part of this help string)
@click.command("api-key", cls=TectonGroup)
def api_key():
    """Interact with Tecton readonly api-keys."""


@api_key.command()
@click.option("--description", default="", help="An optional, human readable description for this API key.")
@click.option(
    "--is-admin",
    is_flag=True,
    default=False,
    help="Whether the API key has admin permissions, generally corresponding to write permissions. Defaults to false.",
)
def create(description, is_admin):
    """Create a new API key."""
    request = CreateApiKeyRequest()
    request.description = description
    request.is_admin = is_admin
    response = metadata_service.instance().CreateApiKey(request)
    printer.safe_print("Save this key - you will not be able to get it again.", file=sys.stderr)
    printer.safe_print(response.key)


@api_key.command()
@click.argument("id", required=True)
def delete(id):
    """Deactivate an API key by its ID."""
    request = DeleteApiKeyRequest()
    try:
        id_proto = IdHelper.from_string(id)
    except:
        printer.safe_print(f"Invalid format for ID")
        sys.exit(1)
    request.id.CopyFrom(id_proto)
    try:
        response = metadata_service.instance().DeleteApiKey(request)
    except tecton._internals.tecton_errors.TectonAPIValidationError as e:
        printer.safe_print(
            f"API key with ID {id} not found. Check `tecton api-key list` to find the IDs of currently active API keys. The key's ID is different from the key's secret value."
        )
        sys.exit(1)
    printer.safe_print("Success")


@api_key.command()
def list():
    """List active API keys."""
    request = ListApiKeysRequest()
    response = metadata_service.instance().ListApiKeys(request)
    for k in response.api_keys:
        printer.safe_print(f"API Key ID: {IdHelper.to_string(k.id)}")
        printer.safe_print(f"Secret Key: {k.obscured_key}")
        printer.safe_print(f"Description: {k.description}")
        printer.safe_print(f"Created by:{k.created_by}")
        printer.safe_print()



from typing import List

from tecton_proto.common.id_pb2 import Id
from tecton_proto.data.fco_pb2 import Fco
from tecton_proto.data.fco_pb2 import FcoContainer as FcoContainerProto
from tecton_proto.data.feature_view_pb2 import FeatureView as FeatureViewProto
from tecton_proto.data.new_transformation_pb2 import NewTransformation as Transformation
from tecton_proto.data.virtual_data_source_pb2 import VirtualDataSource as DataSourceProto
from tecton_spark.id_helper import IdHelper


class FcoWrapper:
    """
    A wrapper class for data.Fco proto, contains convenience accessors
    """

    _fco: Fco

    def __init__(self, fco: Fco):
        self._fco = fco

    @property
    def id(self) -> Id:
        if self._fco.HasField("virtual_data_source"):
            return self._fco.virtual_data_source.virtual_data_source_id
        elif self._fco.HasField("entity"):
            return self._fco.entity.entity_id
        elif self._fco.HasField("new_transformation"):
            return self._fco.new_transformation.transformation_id
        elif self._fco.HasField("feature_view"):
            return self._fco.feature_view.feature_view_id
        elif self._fco.HasField("feature_service"):
            return self._fco.feature_service.feature_service_id
        else:
            raise Exception("unexpected fco type")

    @property
    def inner(self):
        if self._fco.HasField("virtual_data_source"):
            return self._fco.virtual_data_source
        elif self._fco.HasField("entity"):
            return self._fco.entity
        elif self._fco.HasField("new_transformation"):
            return self._fco.new_transformation
        elif self._fco.HasField("feature_view"):
            return self._fco.feature_view
        elif self._fco.HasField("feature_service"):
            return self._fco.feature_service
        else:
            raise Exception("unexpected fco type")


class FcoContainer:
    """
    A wrapper class for FcoContainer proto, contains convenience accessors
    """

    _proto: FcoContainerProto
    _id_to_fco: {}

    def __init__(self, proto: FcoContainerProto):
        assert type(proto) == FcoContainerProto, type(proto)
        self._proto = proto
        self._id_to_fco = {}
        for fco_proto in self._proto.fcos:
            fco_wrapped = FcoWrapper(fco_proto)
            self._id_to_fco[IdHelper.to_string(fco_wrapped.id)] = fco_wrapped

    def get_by_id(self, id: str):
        """
        :return: inner (specific) FCO proto, e.g., Entity, or FeatureView
        """

        return self._id_to_fco[id].inner

    def get_by_ids(self, ids: List[str]):
        """
        :return: inner (specific) FCO protos, e.g., Entities, or FeatureViews
        """

        return [self.get_by_id(id) for id in ids]

    def get_single_root(self):
        """
        :return: inner (specific) FCO proto, pointed by the root_ids or None. Errors if len(root_ids) > 1
        """

        num_root_ids = len(self._proto.root_ids)
        if num_root_ids == 0:
            return None
        elif num_root_ids > 1:
            raise ValueError(f"Expected a single result but got $num_root_ids")
        else:
            id = IdHelper.to_string(self._proto.root_ids[0])
            return self.get_by_id(id)

    def get_root_fcos(self):
        """
        :return: inner (specific) FCO protos, pointed by the root_ids.
        """

        return [self.get_by_id(IdHelper.to_string(id)) for id in self._proto.root_ids]

    def get_all(self):
        """
        :return: all inner (specific) FCO protos
        """

        return [fco.inner for fco in self._id_to_fco.values()]


FCO_CONTAINER_EMTPY = FcoContainer(FcoContainerProto())


def _wrap_data_fco(inner_proto) -> Fco:
    fco = Fco()
    if isinstance(inner_proto, DataSourceProto):
        fco.virtual_data_source.CopyFrom(inner_proto)
    elif isinstance(inner_proto, Transformation):
        fco.new_transformation.CopyFrom(inner_proto)
    elif isinstance(inner_proto, FeatureViewProto):
        fco.feature_view.CopyFrom(inner_proto)
    else:
        raise Exception("Unsupported type " + str(type(inner_proto)))
    return fco


def create_fco_container(fco_protos: List) -> FcoContainerProto:
    proto = FcoContainerProto()
    for inner_fco_proto in fco_protos:
        wrapped_fco_proto = _wrap_data_fco(inner_fco_proto)
        proto.fcos.append(wrapped_fco_proto)
    return FcoContainer(proto)



import json
import time
from typing import List
from typing import Mapping
from typing import Optional
from typing import Union
from urllib.parse import urljoin

import numpy as np
import pandas
import pyspark
import pytimeparse
import requests
from google.protobuf.json_format import MessageToJson

import tecton
from tecton import conf
from tecton import LoggingConfig
from tecton._internals import data_frame_helper
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton._internals.feature_retrieval_internal import find_dependent_feature_set_items
from tecton._internals.sdk_decorators import sdk_public_method
from tecton._internals.utils import filter_internal_columns
from tecton._internals.utils import get_num_dependent_fv
from tecton._internals.utils import infer_timestamp
from tecton._internals.utils import is_live_workspace
from tecton._internals.utils import validate_join_key_types
from tecton._internals.utils import validate_request_data
from tecton._internals.utils import validate_spine_dataframe
from tecton.fco import Fco
from tecton.feature_services.query_helper import _QueryHelper
from tecton.interactive import snowflake_api
from tecton.interactive.data_frame import DataFrame
from tecton.interactive.data_frame import FeatureVector
from tecton.interactive.dataset import Dataset
from tecton.interactive.feature_table import FeatureTable
from tecton.interactive.feature_view import _get_feature_view_by_type
from tecton.interactive.feature_view import FeatureView
from tecton.interactive.feature_view import OnDemandFeatureView
from tecton.tecton_context import TectonContext
from tecton_proto.api.featureservice.feature_service_pb2 import GetFeatureServiceStateRequest
from tecton_proto.data.feature_service_pb2 import FeatureService as FeatureServiceProto
from tecton_proto.metadataservice.metadata_service_pb2 import GetFeatureServiceRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetFeatureServiceSummaryRequest
from tecton_spark.fco_container import FcoContainer
from tecton_spark.feature_set_config import FeatureSetConfig
from tecton_spark.id_helper import IdHelper
from tecton_spark.logger import get_logger
from tecton_spark.transformation import _RequestContext

logger = get_logger("FeatureService")


class FeatureService(Fco):
    """
    FeatureService class.

    FeatureServices are used to serve feature values from :class:`FeatureView`. Users can use FeatureServices
    to make offline requests (e.g. to fetch a training dataset) using the :py:meth:`get_historical_features`
    method, and online requests (e.g. for online serving) using the :py:meth:`get_online_features`
    method. A FeatureService consists of a set of FeatureViews, plus configuration options.

    To get a FeatureService instance, call :py:func:`tecton.get_feature_service`.
    """

    proto: FeatureServiceProto
    _fco_container: FcoContainer
    feature_set_config: FeatureSetConfig
    _uri: str

    _query_helper: _QueryHelper

    def __init__(self):
        """Do not call this directly. Use :py:func:`tecton.get_feature_service`"""

    def __str__(self):
        return f"FeatureService|{self.name}"

    @classmethod
    def _from_proto(cls, feature_service: FeatureServiceProto, fco_container: FcoContainer):
        """
        :param feature_service: FS proto
        :param fco_container: Contains all FS dependencies (transitively), e.g., FVs, Entities, DS-es, Transformations
        """

        from tecton_spark.feature_set_config import FeatureSetConfig

        obj = cls.__new__(cls)

        obj.proto = feature_service
        obj._fco_container = fco_container

        obj.feature_set_config = FeatureSetConfig._from_protos(feature_service.feature_set_items, obj._fco_container)

        # add dependent feature views into the FeatureSetConfig, uniquely per odfv
        # The namespaces of the dependencies have _udf_internal in the name and are filtered out before
        # being returned by TectonContext.execute()
        odfv_ids = set()
        for item in feature_service.feature_set_items:
            if item.HasField("feature_view_id"):
                fv_id = IdHelper.to_string(item.feature_view_id)
                if fv_id in odfv_ids:
                    continue
                odfv_ids.add(fv_id)

                fv = obj._fco_container.get_by_id(fv_id)
                inputs = find_dependent_feature_set_items(
                    obj._fco_container,
                    fv.pipeline.root,
                    visited_inputs={},
                    fv_id=fv_id,
                    workspace_name=obj.workspace,
                )
                obj.feature_set_config._definitions_and_configs = (
                    obj.feature_set_config._definitions_and_configs + inputs
                )
                obj.feature_set_config._ondemand_input_definitions_and_configs = inputs

        return obj

    @classmethod
    def _fco_type_name_singular_snake_case(cls) -> str:
        return "feature_service"

    @classmethod
    def _fco_type_name_plural_snake_case(cls) -> str:
        return "feature_services"

    @property
    def _fco_metadata(self):
        return self.proto.fco_metadata

    @property  # type: ignore
    @sdk_public_method
    def feature_views(self) -> List[Union[FeatureView, FeatureTable]]:
        """
        Returns the Feature Views enclosed in this FeatureService.
        """
        fvs = []
        for d in self.feature_set_config.feature_definitions:
            if d.fv.HasField("feature_table"):
                fvs.append(FeatureTable(d.fv, self._fco_container))
            else:
                fvs.append(_get_feature_view_by_type(d.fv, self._fco_container, d.name))
        return fvs

    @property  # type: ignore
    @sdk_public_method
    def features(self) -> List[str]:
        """
        Returns the features generated by the enclosed feature views.
        """
        return self.feature_set_config.features

    @sdk_public_method
    def query_features(self, join_keys: Mapping[str, Union[np.int_, int, str, bytes]]) -> "tecton.DataFrame":
        """
        [Advanced Feature] Queries the FeatureService with a partial set of join_keys defined in the ``online_serving_index``
        of the included FeatureViews. Returns a Tecton :class:`DataFrame` of all matched records.

        :param join_keys: Query join keys, i.e., a union of join keys in the ``online_serving_index`` of all
            enclosed FeatureViews.
        :return: A Tecton :class:`DataFrame`
        """
        if not is_live_workspace(self.proto.fco_metadata.workspace):
            raise errors.UNSUPPORTED_OPERATION_IN_DEVELOPMENT_WORKSPACE("query_features")
        if not self.proto.online_serving_enabled:
            raise errors.UNSUPPORTED_OPERATION(
                "query_features", "online_serving_enabled was not defined for this Feature Service."
            )
        if not isinstance(join_keys, dict):
            raise errors.INVALID_JOIN_KEYS_TYPE(type(join_keys))

        return _QueryHelper(self.proto.fco_metadata.workspace, feature_service_name=self.name).query_features(join_keys)

    @sdk_public_method
    def get_feature_vector(
        self,
        join_keys: Optional[Mapping[str, Union[int, np.int_, str, bytes]]] = None,
        include_join_keys_in_response: bool = False,
        request_context_map: Optional[Mapping[str, Union[int, np.int_, str, bytes, float]]] = None,
    ) -> FeatureVector:
        """
        Deprecated.
        Returns a single Tecton :class:`FeatureVector` from the Online Store.
        At least one of join_keys or request_context_map is required.

        :param join_keys: Join keys of the enclosed FeatureViews.
        :param include_join_keys_in_response: Whether to include join keys as part of the response FeatureVector.
        :param request_context_map: Dictionary of request context values.

        :return: A :class:`FeatureVector` of the results.
        """
        logger.warning(
            "Deprecated. Use the 'get_online_features' method instead to fetch features from the Online Store for this feature service. "
            + "See the api reference: https://docs.tecton.ai/api-reference/stubs/tecton.interactive.FeatureService.html#tecton.interactive.FeatureService.get_online_features"
        )
        if not self.proto.online_serving_enabled:
            raise errors.UNSUPPORTED_OPERATION(
                "get_feature_vector", "online_serving_enabled was not defined for this Feature Service."
            )
        if not join_keys and not request_context_map:
            raise errors.FS_GET_FEATURE_VECTOR_REQUIRED_ARGS
        if request_context_map is not None and not isinstance(request_context_map, dict):
            raise errors.INVALID_REQUEST_CONTEXT_TYPE(type(request_context_map))

        # do not require join_keys if all fvs are OnDemand with no dependent fvs
        if join_keys is None and not (
            all(
                (
                    isinstance(fv, OnDemandFeatureView)
                    and get_num_dependent_fv(fv._proto.pipeline.root, visited_inputs={}) == 0
                    for fv in self.feature_views
                )
            )
        ):
            raise errors.GET_FEATURE_VECTOR_FS_JOIN_KEYS
        # require request_data if there is an odfv in the FS
        if any((isinstance(fv, OnDemandFeatureView) for fv in self.feature_views)):
            required_request_context_keys = list(self._request_context.arg_to_schema.keys())
            if len(required_request_context_keys) > 0 and request_context_map is None:
                raise errors.GET_FEATURE_VECTOR_FS_NO_REQUEST_DATA(required_request_context_keys)
            validate_request_data(request_context_map, required_request_context_keys, is_read_api_v3=False)

        return self.get_online_features(join_keys, include_join_keys_in_response, request_context_map)

    @sdk_public_method
    def get_online_features(
        self,
        join_keys: Optional[Mapping[str, Union[int, np.int_, str, bytes]]] = None,
        include_join_keys_in_response: bool = False,
        request_data: Optional[Mapping[str, Union[int, np.int_, str, bytes, float]]] = None,
    ) -> FeatureVector:
        """
        Returns a single Tecton :class:`FeatureVector` from the Online Store.
        At least one of join_keys or request_data is required.

        :param join_keys: Join keys of the enclosed FeatureViews.
        :param include_join_keys_in_response: Whether to include join keys as part of the response FeatureVector.
        :param request_data: Dictionary of request context values. Only applicable when the FeatureService contains OnDemandFeatureViews.

        Examples:
            A FeatureService :py:mod:`fs` that contains a BatchFeatureView and StreamFeatureView with join keys :py:mod:`user_id` and :py:mod:`ad_id`.

            1) :py:mod:`fs.get_online_features(join_keys={'user_id': 1, 'ad_id': 'c234'})`
            Fetch the latest features from the online store for user 1 and ad 'c234'.

            2) :py:mod:`fv.get_online_features(join_keys={'user_id': 1, 'ad_id': 'c234'}, include_join_keys_in_respone=True)`
            Fetch the latest features from the online store for user 1 and ad id 'c234'. Include the join key information (user_id=1, ad_id='c234') in the returned FeatureVector.

            A FeatureService :py:mod:`fs_on_demand` that contains only OnDemandFeatureViews and expects request time data for key :py:mod:`amount`.

            | The request time data is defined in the feature definition as such:
            | `request_schema = StructType()`
            | `request_schema.add(StructField('amount', DoubleType()))`
            | `transaction_request = RequestDataSource(request_schema=request_schema)`

            1) :py:mod:`fs.get_online_features(request_data={'amount': 30})`
            Fetch the latest features from the online store with amount = 30.

            A FeatureService :py:mod:`fs_all` that contains feature views of all types with join key :py:mod:`user_id` and expects request time data for key :py:mod:`amount`.

            1) :py:mod:`fs.get_online_features(join_keys={'user_id': 1}, request_data={'amount': 30})`
            Fetch the latest features from the online store for user 1 with amount = 30.

        :return: A :class:`FeatureVector` of the results.
        """
        if not self.proto.online_serving_enabled:
            raise errors.UNSUPPORTED_OPERATION(
                "get_online_features", "online_serving_enabled was not defined for this Feature Service."
            )
        if not join_keys and not request_data:
            raise errors.FS_GET_ONLINE_FEATURES_REQUIRED_ARGS
        if join_keys is not None:
            validate_join_key_types(join_keys)
        if request_data is not None and not isinstance(request_data, dict):
            raise errors.INVALID_REQUEST_DATA_TYPE(type(request_data))

        # do not require join_keys if all fvs are OnDemand with no dependent fvs
        if join_keys is None and not (
            all(
                (
                    isinstance(fv, OnDemandFeatureView)
                    and get_num_dependent_fv(fv._proto.pipeline.root, visited_inputs={}) == 0
                    for fv in self.feature_views
                )
            )
        ):
            raise errors.GET_ONLINE_FEATURES_FS_JOIN_KEYS

        # require request_data param if there are request context keys
        required_request_context_keys = list(self._request_context.arg_to_schema.keys())
        if len(required_request_context_keys) > 0 and request_data is None:
            raise errors.GET_ONLINE_FEATURES_FS_NO_REQUEST_DATA(required_request_context_keys)
        validate_request_data(request_data, required_request_context_keys)

        return _QueryHelper(self.proto.fco_metadata.workspace, feature_service_name=self.name).get_feature_vector(
            join_keys or {},
            include_join_keys_in_response,
            request_data or {},
            self._request_context,
        )

    @classmethod
    def _get_feature_dataframe_internal(
        cls,
        feature_set_config: FeatureSetConfig,
        spine: Union[str, dict, "pyspark.sql.DataFrame", list, "pandas.DataFrame", "pyspark.RDD", None],
        timestamp_key: str,
        include_feature_view_timestamp_columns=False,
        use_materialized_data: bool = True,
    ) -> DataFrame:
        """
        Creates a feature vector dataframe for the specified entities, including the entities' features.

        :param feature_set_config: FeatureSetConfig of a feature service or collection of FVs
        :param spine: SQL string that fetches the entity ids for which the features will be generated or a dataframe.
        :param timestamp_key: Name of the time column in spine. The column must be of type Spark timestamp.
        :param include_feature_view_timestamp_columns: (Optional) Include timestamp columns for every individual feature definition.
        :param use_materialized_data: (Optional) Use materialized data if materialization is enabled.
        :return: A Tecton DataFrame.
        """
        from tecton.tecton_context import TectonContext

        tc = TectonContext.get_instance()
        return tc.execute(
            spine,
            feature_set_config=feature_set_config,
            timestamp_key=timestamp_key,
            include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
            use_materialized_data=use_materialized_data,
        )

    @sdk_public_method
    def get_feature_dataframe(
        self,
        spine: Union[pyspark.sql.dataframe.DataFrame, pandas.DataFrame, None],
        timestamp_key: Optional[str] = None,
        include_feature_view_timestamp_columns: bool = False,
        use_materialized_data: bool = True,
        save: bool = None,
        save_as: str = None,
    ) -> DataFrame:
        """
        Deprecated.
        Fetch a :class:`DataFrame` of feature values from this FeatureService.

        This method will return feature values for each row provided in the spine DataFrame. The feature values
        returned by this method will respect the timestamp provided in the timestamp column of the Spine DataFrame.

        This method fetches features from the Offline Store. If ``use_materialized_data==False``, feature values
        will instead be computed on the fly from raw data.

        :param spine: A dataframe of join keys and timestamps that specify which feature values to fetch.
        :param timestamp_key: Name of the time column in the spine dataframe. This column must be of type Spark timestamp.
        :param include_feature_view_timestamp_columns: (Optional) Include timestamp columns for each FeatureView in the FeatureService.
        :param use_materialized_data: (Optional) Use materialized data if materialization is enabled.
        :param save: (Optional) set to True to persist DataFrame as a Dataset object
        :param save_as: (Optional) name to save the DataFrame as. Not applicable when save=False.
            If unspecified and save=True, a name will be generated.
        :return: A Tecton :class:`DataFrame`.
        """
        logger.warning(
            "Deprecated. Use the 'get_historical_features' method instead to view historical values for this feature service."
            + "See the api reference: https://docs.tecton.ai/api-reference/stubs/tecton.interactive.FeatureService.html#tecton.interactive.FeatureService.get_historical_features"
        )
        # handling case where spine is None (note we don't allow the spine to be none in fs.get_historical_features)
        if spine is None:
            TectonContext.validate_spine_type(spine)
            if timestamp_key is None:
                keys = [fv.timestamp_key for fv in self.feature_views if not fv.is_on_demand]
                if len(keys) > 0:
                    if not all([key == keys[0] for key in keys]):
                        raise errors.FS_AMBIGUOUS_TIMESTAMP_KEY(keys)
                    timestamp_key = keys[0]

            df = self._get_feature_dataframe_internal(
                self.feature_set_config,
                spine,
                timestamp_key=timestamp_key,
                include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
                use_materialized_data=use_materialized_data,
            )
            if save or save_as is not None:
                return Dataset._create(
                    df=df,
                    save_as=save_as,
                    workspace=self.workspace,
                    feature_service_id=self.id,
                    timestamp_key=timestamp_key,
                )
            return df
        # check here instead of get_historical_features so error includes correct param name
        if (
            any(isinstance(fv, FeatureTable) for fv in self.feature_views)
            and is_live_workspace(self.workspace)
            and not use_materialized_data
        ):
            raise errors.USE_MATERIALIZED_DATA_WITH_FT

        return self.get_historical_features(
            spine, timestamp_key, include_feature_view_timestamp_columns, not use_materialized_data, save, save_as
        )

    @sdk_public_method
    def get_historical_features(
        self,
        spine: Union[pyspark.sql.dataframe.DataFrame, pandas.DataFrame, DataFrame, str],
        timestamp_key: Optional[str] = None,
        include_feature_view_timestamp_columns: bool = False,
        from_source: bool = False,
        save: bool = False,
        save_as: Optional[str] = None,
    ) -> DataFrame:
        """
        Fetch a :class:`DataFrame` of feature values from this FeatureService.

        This method will return feature values for each row provided in the spine DataFrame. The feature values
        returned by this method will respect the timestamp provided in the timestamp column of the spine DataFrame.

        This method fetches features from the Offline Store. If ``from_source=True``, feature values
        will instead be computed on the fly from raw data.

        :param spine: A dataframe of possible join keys, request data keys, and timestamps that specify which feature values to fetch.To distinguish
            between spine columns and feature columns, feature columns are labeled as `feature_view_name.feature_name`
            in the returned DataFrame.
        :type spine: Union[pyspark.sql.DataFrame, pandas.DataFrame, tecton.DataFrame]
        :param timestamp_key: (Optional) Name of the time column in the spine DataFrame.
            This method will fetch the latest features computed before the specified timestamps in this column.
            Not applicable if the FeatureService stricly contains OnDemandFeatureViews with no feature view dependencies.
        :type timestamp_key: str
        :param include_feature_view_timestamp_columns: (Optional) Whether to include timestamp columns for each FeatureView in the FeatureService. Default is False.
        :type include_feature_view_timestamp_columns: bool
        :param from_source: (Optional) Whether feature values should be recomputed from the original data source.
            If False, we will read the values from the materialized Offline store.
        :type from_source: bool
        :param save: (Optional) Whether to persistthe  DataFrame as a Dataset object. Default is False.
        :type save: bool
        :param save_as: (Optional) Name to save the DataFrame as.
            If unspecified and save=True, a name will be generated.
        :type save_as: str

        Examples:
            A FeatureService :py:mod:`fs` that contains a BatchFeatureView and StreamFeatureView with join keys :py:mod:`user_id` and :py:mod:`ad_id`.

            1) :py:mod:`fs.get_historical_features(spine)` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'ad_id': ['a234', 'b256', 'c9102'], 'date': [datetime(...), datetime(...), datetime(...)]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the specified timestamps and ad ids in the spine.

            2) :py:mod:`fs.get_historical_features(spine, save_as='my_dataset)` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'ad_id': ['a234', 'b256', 'c9102'], 'date': [datetime(...), datetime(...), datetime(...)]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the specified timestamps and ad ids in the spine. Save the DataFrame as dataset with the name :py:mod:`my_dataset`.

            3) :py:mod:`fv.get_historical_features(spine, timestamp_key='date_1')` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'ad_id': ['a234', 'b256', 'c9102'],
            'date_1': [datetime(...), ...], 'date_2': [datetime(...), ...]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the ad ids and specified timestamps in the :py:mod:`date_1` column in the spine.

            A FeatureService :py:mod:`fs_on_demand` that contains only OnDemandFeatureViews and expects request time data for the key :py:mod:`amount`.

            | The request time data is defined in the feature definition as such:
            | `request_schema = StructType()`
            | `request_schema.add(StructField('amount', DoubleType()))`
            | `transaction_request = RequestDataSource(request_schema=request_schema)`

            1) :py:mod:`fs_on_demand.get_historical_features(spine)` where :py:mod:`spine=pandas.Dataframe({'amount': [30, 50, 10000]})`
            Fetch historical features from the offline store with request data inputs 30, 50, and 10000.

            A FeatureService :py:mod:`fs_all` that contains feature views of all types with join key 'user_id' and expects request time data for the key :py:mod:`amount`.

            1) :py:mod:`fs_all.get_historical_features(spine)` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'amount': [30, 50, 10000], 'date': [datetime(...), datetime(...), datetime(...)]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the specified timestamps and request data inputs in the spine.

        :return: A Tecton :class:`DataFrame`.
        """
        if conf.get_or_none("ALPHA_SNOWFLAKE_COMPUTE_ENABLED"):
            return snowflake_api.get_historical_features(
                spine=spine,
                timestamp_key=timestamp_key,
                include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
                from_source=from_source,
                save=save,
                save_as=save_as,
                feature_set_config=self.feature_set_config,
            )
        # str is only supported for Snowflake
        if isinstance(spine, str):
            raise TypeError(
                "spine must be one of (pyspark.sql.dataframe.DataFrame, pandas.core.frame.DataFrame, tecton.interactive.data_frame.DataFrame); got str instead"
            )

        if any(isinstance(fv, FeatureTable) for fv in self.feature_views):
            if not is_live_workspace(self.workspace):
                raise errors.FS_WITH_FT_DEVELOPMENT_WORKSPACE
            elif from_source:
                raise errors.FROM_SOURCE_WITH_FT

        spark = TectonContext.get_instance()._spark
        if isinstance(spine, pandas.DataFrame):
            spine = spark.createDataFrame(spine)
        elif isinstance(spine, DataFrame):
            spine = spine.to_spark()

        timestamp_required = any(
            [fv._should_infer_timestamp_of_spine(timestamp_key, spine) for fv in self.feature_views]
        )

        if timestamp_required:
            timestamp_key = timestamp_key or infer_timestamp(spine)
        validate_spine_dataframe(spine, timestamp_key, list(self._request_context.arg_to_schema.keys()))

        df = data_frame_helper.get_features_for_spine(
            spark,
            spine,
            self.feature_set_config,
            timestamp_key=timestamp_key,
            from_source=from_source,
            include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
        )
        df = filter_internal_columns(df)
        df = DataFrame._create(df)
        if save or save_as is not None:
            return Dataset._create(
                df=df,
                save_as=save_as,
                workspace=self.workspace,
                feature_service_id=self.id,
                spine=spine,
                timestamp_key=timestamp_key,
            )
        return df

    @sdk_public_method
    def wait_until_ready(self, timeout="15m", wait_for_materialization=True, verbose=False):
        """Blocks until the service is ready to serve real-time requests.

        The FeatureService is considered ready once every FeatureView that has been added to it
        has had at least once successful materialization run.

        :param timeout: Timeout string.
        :param wait_for_materialization: If False, does not wait for batch materialization to complete.
        """
        if not is_live_workspace(self.proto.fco_metadata.workspace):
            raise errors.UNSUPPORTED_OPERATION_IN_DEVELOPMENT_WORKSPACE("wait_until_ready")

        timeout_seconds = pytimeparse.parse(timeout)
        deadline = time.time() + timeout_seconds

        has_been_not_ready = False
        while True:
            request = GetFeatureServiceStateRequest()
            request.feature_service_locator.feature_service_name = self.name
            request.feature_service_locator.workspace_name = self.proto.fco_metadata.workspace
            http_response = requests.post(
                urljoin(conf.get_or_raise("FEATURE_SERVICE") + "/", "v1/feature-service/get-feature-service-state"),
                data=MessageToJson(request),
                headers=_QueryHelper(
                    self.proto.fco_metadata.workspace, feature_service_name=self.name
                )._prepare_headers(),
            )
            details = http_response.json()
            if http_response.status_code == 404:
                # FeatureService is not ready to serve
                if verbose:
                    logger.info(f" Waiting for FeatureService to be ready to serve ({details['message']})")
                else:
                    logger.info(f" Waiting for FeatureService to be ready to serve")
            elif http_response.status_code == 200:
                # FeatureService is ready
                if verbose:
                    logger.info(f"wait_until_ready: Ready! Response={http_response.text}")
                else:
                    logger.info(f"wait_until_ready: Ready!")
                # Extra wait time due to different FS hosts being potentially out-of-sync in picking up the latest state
                if has_been_not_ready:
                    time.sleep(20)
                return
            else:
                http_response.raise_for_status()
                return
            if time.time() > deadline:
                logger.info(f"wait_until_ready: Response={http_response.text}")
                raise TimeoutError()
            has_been_not_ready = True
            time.sleep(10)
            continue

    @sdk_public_method
    def summary(self) -> Displayable:
        """
        Returns various information about this FeatureService, including the most critical metadata such
        as the FeatureService's name, owner, features, etc.
        """
        return Displayable.from_items(headings=["Property", "Value"], items=self._summary_items())

    @property  # type: ignore
    @sdk_public_method
    def id(self) -> str:
        """
        Legacy attribute. Use name to refer to a FeatureService.
        """
        return IdHelper.to_string(self.proto.feature_service_id)

    def _summary_items(self):
        request = GetFeatureServiceSummaryRequest()
        request.feature_service_id.CopyFrom(self.proto.feature_service_id)
        request.workspace = self.workspace
        response = metadata_service.instance().GetFeatureServiceSummary(request)
        items_map = {}
        summary_items = []
        for item in response.general_items:
            items_map[item.key] = item
            if item.display_name:
                summary_items.append((item.display_name, item.multi_values if item.multi_values else item.value))

        if "curlEndpoint" in items_map and "curlParamsJson" in items_map:
            api_service = conf.get_or_none("API_SERVICE")
            assert api_service is not None, "API_SERVICE must be configured"
            if "localhost" not in api_service and "ingress" not in api_service:
                assert api_service.endswith(
                    "/api"
                ), "API_SERVICE should look like https://<deployment-name>.tecton.ai/api"
            service_url = urljoin(api_service, items_map["curlEndpoint"].value)
            curl_params_json = json.dumps(json.loads(items_map["curlParamsJson"].value), indent=2)
            curl_header = items_map["curlHeader"].value
            curl_str = "curl -X POST " + service_url + '\\\n-H "' + curl_header + "\"-d\\\n'" + curl_params_json + "'"
            summary_items.append(("Example Curl", curl_str))
        return summary_items

    @property
    def _request_context(self):
        merged_context = _RequestContext({})
        for fv in self.feature_views:
            if isinstance(fv, OnDemandFeatureView):
                merged_context._merge(fv._request_context)
        return merged_context

    @property
    def logging(self) -> Optional["LoggingConfig"]:
        """
        Returns the logging configuration of this FeatureService.
        """
        return LoggingConfig._from_proto(self.proto.logging)


@sdk_public_method
def get_feature_service(name: str, workspace_name: Optional[str] = None) -> FeatureService:
    """
    Fetch an existing :class:`FeatureService` by name.

    :param name: Name or string ID of the :class:`FeatureService`.
    :return: An instance of :class:`FeatureService`.
    """

    if workspace_name == None:
        logger.warning(
            "`tecton.get_feature_service('<name>')` is deprecated. Please use `tecton.get_workspace('<workspace_name>').get_feature_service('<name>')` instead."
        )

    request = GetFeatureServiceRequest()
    request.service_reference = name
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    request.disable_legacy_response = True
    response = metadata_service.instance().GetFeatureService(request)
    fco_container = FcoContainer(response.fco_container)
    fs_proto = fco_container.get_single_root()

    if not fs_proto:
        raise errors.FCO_NOT_FOUND(FeatureService, name)

    return FeatureService._from_proto(fs_proto, fco_container)



from typing import Callable
from typing import List
from typing import Optional

from tecton.data_sources.base_data_source_config import BaseBatchDSConfig
from tecton.data_sources.datetime_partition_column import DatetimePartitionColumn
from tecton_proto.args import data_source_pb2
from tecton_proto.args import virtual_data_source_pb2
from tecton_spark import function_serialization


class HiveDSConfig(BaseBatchDSConfig):
    """
    Configuration used to reference a Hive table.

    The HiveDSConfig class is used to create a reference to a Hive Table.

    This class used as an input to a :class:`BatchDataSource`'s parameter ``batch_ds_config``. This class is not
    a Tecton Object: it is a grouping of parameters. Declaring this class alone will not register a data source.
    Instead, declare as part of ``BatchDataSource`` that takes this configuration class instance as a parameter.
    """

    def __init__(
        self,
        table: str,
        database: str,
        date_partition_column: str = None,
        timestamp_column_name: str = None,
        timestamp_format: str = None,
        skip_validation=False,
        datetime_partition_columns: List[DatetimePartitionColumn] = None,
        raw_batch_translator=None,
    ):
        """
        Instantiates a new HiveDSConfig.

        :param table: A table registered in Hive MetaStore.
        :param database: A database registered in Hive MetaStore.
        :param date_partition_column: (Optional) Partition column name in case the raw data is partitioned by date, otherwise None.
        :param datetime_partition_columns: (Optional) List of DatetimePartitionColumn the raw data is partitioned by, otherwise None.
        :param timestamp_column_name: Name of timestamp column.
        :param timestamp_format: (Optional) Format of string-encoded timestamp column (e.g. "yyyy-MM-dd'T'hh:mm:ss.SSS'Z'").
                                 If the timestamp string cannot be parsed with this format, Tecton will fallback and attempt to
                                 use the default timestamp parser.
        :param raw_batch_translator: Python user defined function f(DataFrame) -> DataFrame that takes in raw
                                     PySpark data source DataFrame and translates it to the DataFrame to be
                                     consumed by the Feature View. See an example of
                                     raw_batch_translator in the `User Guide`_.

        :return: A HiveDSConfig class instance.

        .. _User Guide: https://docs.tecton.ai/v2/overviews/framework/data_sources.html

        Example of a HiveDSConfig declaration:

        .. code-block:: python

            from tecton import HiveDSConfig
            import pyspark

            def convert_temperature(df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
                from pyspark.sql.functions import udf,col
                from pyspark.sql.types import DoubleType

                # Convert the incoming PySpark DataFrame temperature Celsius to Fahrenheit
                udf_convert = udf(lambda x: x * 1.8 + 32.0, DoubleType())
                converted_df = df.withColumn("Fahrenheit", udf_convert(col("Temperature"))).drop("Temperature")
                return converted_df

            # declare a HiveDSConfig instance, which can be used as a parameter to a BatchDataSource
            batch_ds_config=HiveDSConfig(database='global_temperatures',
                                        table='us_cities',
                                        timestamp_column_name='timestamp',
                                        raw_batch_translator=convert_temperature)

        """
        self._args = prepare_hive_ds_args(
            table=table,
            database=database,
            date_partition_column=date_partition_column,
            timestamp_column_name=timestamp_column_name,
            timestamp_format=timestamp_format,
            datetime_partition_columns=datetime_partition_columns,
            raw_batch_translator=raw_batch_translator,
        )

    def _merge_batch_args(self, data_source_args: virtual_data_source_pb2.VirtualDataSourceArgs):
        data_source_args.hive_ds_config.CopyFrom(self._args)


def prepare_hive_ds_args(
    *,
    table: str,
    database: str,
    date_partition_column: Optional[str],
    timestamp_column_name: Optional[str],
    timestamp_format: Optional[str],
    datetime_partition_columns: Optional[List[DatetimePartitionColumn]],
    raw_batch_translator: Optional[Callable],
):
    args = data_source_pb2.HiveDataSourceArgs()
    args.table = table
    args.database = database
    if date_partition_column:
        args.date_partition_column = date_partition_column
    if timestamp_column_name:
        args.timestamp_column_name = timestamp_column_name
    if timestamp_format:
        args.timestamp_format = timestamp_format

    if datetime_partition_columns:
        for column in datetime_partition_columns:
            column_args = data_source_pb2.DatetimePartitionColumnArgs()
            column_args.column_name = column.column_name
            column_args.datepart = column.datepart
            column_args.zero_padded = column.zero_padded
            args.datetime_partition_columns.append(column_args)
    if raw_batch_translator is not None:
        args.raw_batch_translator.CopyFrom(function_serialization.to_proto(raw_batch_translator))

    return args



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/entity.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import basic_info_pb2 as tecton__proto_dot_args_dot_basic__info__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.args import diff_options_pb2 as tecton__proto_dot_args_dot_diff__options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/args/entity.proto',
  package='tecton_proto.args',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.argsP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1etecton_proto/args/entity.proto\x12\x11tecton_proto.args\x1a\"tecton_proto/args/basic_info.proto\x1a\x1ctecton_proto/common/id.proto\x1a$tecton_proto/args/diff_options.proto\"\x98\x01\n\nEntityArgs\x12\x34\n\tentity_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x08\x65ntityId\x12\x37\n\x04info\x18\x02 \x01(\x0b\x32\x1c.tecton_proto.args.BasicInfoB\x05\x92M\x02\x10\x01R\x04info\x12\x1b\n\tjoin_keys\x18\x03 \x03(\tR\x08joinKeysB\x13\n\x0f\x63om.tecton.argsP\x01'
  ,
  dependencies=[tecton__proto_dot_args_dot_basic__info__pb2.DESCRIPTOR,tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_diff__options__pb2.DESCRIPTOR,])




_ENTITYARGS = _descriptor.Descriptor(
  name='EntityArgs',
  full_name='tecton_proto.args.EntityArgs',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entity_id', full_name='tecton_proto.args.EntityArgs.entity_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='entityId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='info', full_name='tecton_proto.args.EntityArgs.info', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\222M\002\020\001', json_name='info', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='join_keys', full_name='tecton_proto.args.EntityArgs.join_keys', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='joinKeys', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=158,
  serialized_end=310,
)

_ENTITYARGS.fields_by_name['entity_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_ENTITYARGS.fields_by_name['info'].message_type = tecton__proto_dot_args_dot_basic__info__pb2._BASICINFO
DESCRIPTOR.message_types_by_name['EntityArgs'] = _ENTITYARGS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EntityArgs = _reflection.GeneratedProtocolMessageType('EntityArgs', (_message.Message,), {
  'DESCRIPTOR' : _ENTITYARGS,
  '__module__' : 'tecton_proto.args.entity_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.EntityArgs)
  })
_sym_db.RegisterMessage(EntityArgs)


DESCRIPTOR._options = None
_ENTITYARGS.fields_by_name['info']._options = None
# @@protoc_insertion_point(module_scope)



from datetime import datetime
from datetime import timedelta
from typing import Dict
from typing import Union

from tecton._internals.display import Displayable
from tecton._internals.sdk_decorators import sdk_public_method
from tecton._internals.utils import can_be_stale
from tecton._internals.utils import format_freshness_table
from tecton._internals.utils import get_all_freshness


@sdk_public_method
def get_feature_freshness(workspace_name, to_dict: bool = False) -> Union[Dict[str, Dict], Displayable, str]:
    """
    Fetch freshness status for features

    :param: workspace_name: workspace to fetch freshness statuses of features from.
    :param: to_dict: return data as a Python dictionary. Otherwise, data is printed as a table.

    :return: Table or dictionary of freshness statuses for all features
    """
    # TODO: use GetAllFeatureFreshnessRequest once we implement Chronosphere based API.
    freshness_statuses = get_all_freshness(workspace_name)

    if to_dict:
        to_return = {}
        for ff_proto in freshness_statuses:
            to_return[ff_proto.feature_view_name] = {
                "materialization_enabled": ff_proto.materialization_enabled,
                "is_stale": ff_proto.is_stale if can_be_stale(ff_proto) else None,
                "freshness": timedelta(seconds=ff_proto.freshness.seconds) if can_be_stale(ff_proto) else None,
                "expected_freshness": timedelta(seconds=ff_proto.expected_freshness.seconds)
                if can_be_stale(ff_proto)
                else None,
                "created": datetime.fromtimestamp(ff_proto.created_at.seconds),
                "is_stream": ff_proto.is_stream,
            }
        return to_return

    if len(freshness_statuses) == 0:
        return "No Feature Views found in this workspace"
    return format_freshness_table(freshness_statuses)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/diff_options.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/args/diff_options.proto',
  package='tecton_proto.args',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.argsP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n$tecton_proto/args/diff_options.proto\x12\x11tecton_proto.args\x1a google/protobuf/descriptor.proto\"e\n\x0b\x44iffOptions\x12\x39\n\x06update\x18\x01 \x01(\x0e\x32!.tecton_proto.args.UpdateStrategyR\x06update\x12\x1b\n\thide_path\x18\x02 \x01(\x08R\x08hidePath*{\n\x0eUpdateStrategy\x12\x0c\n\x08RECREATE\x10\x00\x12\x0b\n\x07INPLACE\x10\x01\x12\x12\n\x0eINPLACE_ON_ADD\x10\x03\x12\x15\n\x11INPLACE_ON_REMOVE\x10\x04\x12\x0b\n\x07PASSIVE\x10\x05\x12\x16\n\x12INPLACE_BREAKGLASS\x10\x06:a\n\x0c\x64iff_options\x12\x1d.google.protobuf.FieldOptions\x18\xd2\t \x01(\x0b\x32\x1e.tecton_proto.args.DiffOptionsR\x0b\x64iffOptionsB\x13\n\x0f\x63om.tecton.argsP\x01'
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,])

_UPDATESTRATEGY = _descriptor.EnumDescriptor(
  name='UpdateStrategy',
  full_name='tecton_proto.args.UpdateStrategy',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RECREATE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INPLACE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INPLACE_ON_ADD', index=2, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INPLACE_ON_REMOVE', index=3, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PASSIVE', index=4, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INPLACE_BREAKGLASS', index=5, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=196,
  serialized_end=319,
)
_sym_db.RegisterEnumDescriptor(_UPDATESTRATEGY)

UpdateStrategy = enum_type_wrapper.EnumTypeWrapper(_UPDATESTRATEGY)
RECREATE = 0
INPLACE = 1
INPLACE_ON_ADD = 3
INPLACE_ON_REMOVE = 4
PASSIVE = 5
INPLACE_BREAKGLASS = 6

DIFF_OPTIONS_FIELD_NUMBER = 1234
diff_options = _descriptor.FieldDescriptor(
  name='diff_options', full_name='tecton_proto.args.diff_options', index=0,
  number=1234, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, json_name='diffOptions', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)


_DIFFOPTIONS = _descriptor.Descriptor(
  name='DiffOptions',
  full_name='tecton_proto.args.DiffOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='update', full_name='tecton_proto.args.DiffOptions.update', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='update', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hide_path', full_name='tecton_proto.args.DiffOptions.hide_path', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='hidePath', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=93,
  serialized_end=194,
)

_DIFFOPTIONS.fields_by_name['update'].enum_type = _UPDATESTRATEGY
DESCRIPTOR.message_types_by_name['DiffOptions'] = _DIFFOPTIONS
DESCRIPTOR.enum_types_by_name['UpdateStrategy'] = _UPDATESTRATEGY
DESCRIPTOR.extensions_by_name['diff_options'] = diff_options
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DiffOptions = _reflection.GeneratedProtocolMessageType('DiffOptions', (_message.Message,), {
  'DESCRIPTOR' : _DIFFOPTIONS,
  '__module__' : 'tecton_proto.args.diff_options_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.DiffOptions)
  })
_sym_db.RegisterMessage(DiffOptions)

diff_options.message_type = _DIFFOPTIONS
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(diff_options)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



def set_credentials(tecton_api_key: str) -> None:
    """
    Explicitly override tecton credentials settings.

    Typically, Tecton credentials are set in environment variables, but if your
    Tecton SDK setup requires another type of setup, you can use this function
    to set the Tecton API Key secret during an interactive Python session.

    :param tecton_api_key: Tecton API Key
    """
    # Import this lazily so we don't trigger initialization that happens on importing conf module.
    from tecton import conf

    conf.set("TECTON_API_KEY", tecton_api_key)



import datetime
from typing import Union

import pendulum
from pyspark.sql import Column
from pyspark.sql import functions
from pyspark.sql.functions import expr

from tecton_proto.data.feature_store_pb2 import FeatureStoreFormatVersion
from tecton_spark.time_utils import nanos_to_seconds


def subtract_seconds_from_timestamp(
    timestamp: Union[Column, datetime.datetime], delta_seconds: int
) -> Union[Column, datetime.datetime]:
    """
    Subtract seconds from timestamp
    Timestamp can be in Column[Timestamp] or just Timestamp
    :param timestamp: seconds
    :param delta_seconds: seconds
    :return:
    """
    td = datetime.timedelta(seconds=delta_seconds)
    if isinstance(timestamp, Column):
        return timestamp - expr(f"INTERVAL {td.days} DAYS") - expr(f"INTERVAL {td.seconds} SECONDS")
    else:
        return timestamp - td


def add_seconds_to_timestamp(
    timestamp: Union[Column, datetime.datetime], delta_seconds: int
) -> Union[Column, datetime.datetime]:
    """
    Add seconds to timestamp
    Timestamp can be in Column[Timestamp] or just Timestamp
    :param timestamp:  seconds
    :param delta_seconds: seconds
    :return:
    """
    if isinstance(timestamp, Column):
        td = datetime.timedelta(seconds=delta_seconds)
        return timestamp + expr(f"INTERVAL {td.days} DAYS") + expr(f"INTERVAL {td.seconds} SECONDS")
    else:
        td = datetime.timedelta(seconds=delta_seconds)
        return timestamp + td


def convert_timestamp_to_epoch(timestamp: Union[Column, datetime.datetime], version: int) -> Union[Column, int]:
    """
    Convert timestamp to epoch
    Timestamp can be in Column[Timestamp] or just Timestamp
    V0 Return Epoch in seconds
    V1 Return Epoch in nanoseconds
    :param timestamp: Datetime / Datetime column
    :param version: Feature Store Format Version
    :return:
    """
    if version == FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return get_timestamp_in_seconds(timestamp)
    else:
        return get_timestamp_in_nanos(timestamp)


def convert_epoch_to_datetime(epoch: Union[Column, int], version: int) -> Union[Column, pendulum.datetime]:
    """
    Convert epoch to datetime
    V0 epoch is in seconds
    V1 epoch is in nanoseconds -> convert to seconds
    :param epoch: Epoch based on version
    :param version: Feature Store Format Version
    :return:
    """
    if isinstance(epoch, Column):
        return convert_epoch_to_timestamp_column(epoch, version)
    else:
        epoch_float = float(epoch)
        if version >= FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_TIME_NANOSECONDS:
            epoch_float = nanos_to_seconds(epoch)
        return pendulum.from_timestamp(epoch_float)


def get_timestamp_in_seconds(timestamp: Union[Column, datetime.datetime]) -> Union[Column, int]:
    """
    Converts a given timestamp to epoch total seconds
    Timestamp can be in Column[Timestamp] or just Timestamp
    If its a column we need to use a UDF.
    :param timestamp: Datetime / Datetime column
    :return: Timestamp / Timestamp column
    """
    if isinstance(timestamp, Column):
        return timestamp.cast("int")
    else:
        return int(timestamp.timestamp())


def get_timestamp_in_nanos(timestamp: Union[Column, datetime.datetime]) -> Union[Column, int]:
    """
    Converts a given timestamp to epoch total nano seconds
    Timestamp could be a column of timestamp type or
    an actual timestamp. If its a column we need to use
    a UDF.
    :param timestamp: Datetime / Datetime column
    :return: Timestamp / Timestamp column
    """
    if isinstance(timestamp, Column):
        return ((timestamp.cast("double") * 1e6).cast("long")) * 1000
    else:
        return (int(timestamp.timestamp() * 1e6)) * 1000


def convert_epoch_to_timestamp_column(epoch: int, version: int) -> Column:
    """
    Converts an epoch column to a timestamp column
    :param epoch: Epoch Column [V0 : Seconds, V1 : Nanoseconds]
    :param version: Feature Store Format Version
    :return:
    """
    if version == FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return functions.to_timestamp(epoch)
    else:
        return functions.to_timestamp(epoch / functions.lit(1e9))



from typing import Dict
from typing import List
from typing import Union

from pyspark.sql.types import ArrayType
from pyspark.sql.types import AtomicType
from pyspark.sql.types import BooleanType
from pyspark.sql.types import DoubleType
from pyspark.sql.types import FloatType
from pyspark.sql.types import LongType
from pyspark.sql.types import StringType
from pyspark.sql.types import StructType

from tecton_proto.args.pipeline_pb2 import RequestContext as RequestContextProto
from tecton_proto.data.new_transformation_pb2 import NewTransformation as TransformationDataProto
from tecton_proto.data.virtual_data_source_pb2 import VirtualDataSource as VirtualDataSourceProto
from tecton_spark import errors
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper

InputType = Union["VirtualDataSourceProto", "TransformationDataProto"]
TRANSFORMATION_TEMP_VIEW_PREFIX = "_tecton_transformation_"

VALID_FEATURE_SERVER_TYPES = [
    LongType(),
    DoubleType(),
    StringType(),
    BooleanType(),
    ArrayType(LongType()),
    ArrayType(FloatType()),
    ArrayType(DoubleType()),
    ArrayType(StringType()),
]
VALID_FEATURE_SERVER_TYPES_ERROR_STR = ", ".join([str(t) for t in VALID_FEATURE_SERVER_TYPES])


class _RequestContext:
    """
    Wrapper around the RequestContext proto object (that is part of the RequestDataSourceNode in the FV pipeline).
    """

    def __init__(self, schema: Dict[str, AtomicType]):
        """
        Creates a _RequestContext object.

        :param schema: Dictionary mapping string -> data types. Supported types are LongType, DoubleType, StringType, BooleanType.
        """
        for field, typ in schema.items():
            if typ not in VALID_FEATURE_SERVER_TYPES:
                raise errors.TectonValidationError(
                    f"RequestContext schema type {typ} for field '{field}' not supported. Expected one of {VALID_FEATURE_SERVER_TYPES_ERROR_STR}"
                )

        self.arg_to_schema = schema

    def _to_proto(self) -> RequestContextProto:
        s = StructType()
        for field, typ in self.arg_to_schema.items():
            s.add(field, typ)
        wrapper = SparkSchemaWrapper(s)
        return RequestContextProto(schema=wrapper.to_proto())

    def _set_schema_field_order(self, schema_field_order: List[str]):
        """
        Forces an ordering for schemas with multiple fields.

        :param schema_field_order: Ordered list of schema fields.
        """
        assert set(schema_field_order) == set(self.arg_to_schema.keys()) and len(schema_field_order) == len(
            self.arg_to_schema.keys()
        ), f"Schema ordering fields {schema_field_order} must contain the same elements as schema dictionary keys {list(self.arg_to_schema.keys())}."
        new_schema = {}
        for field in schema_field_order:
            new_schema[field] = self.arg_to_schema[field]
        self.arg_to_schema = new_schema

    @classmethod
    def _from_proto(cls, proto: RequestContextProto):
        wrapper = SparkSchemaWrapper.from_proto(proto.schema)
        schema_dict = {field: typ for field, typ in wrapper.column_name_types()}
        return _RequestContext(schema=schema_dict)

    def _merge(self, other):
        for field in other.arg_to_schema:
            if field in self.arg_to_schema:
                # should not happen
                assert (
                    self.arg_to_schema[field] == other.arg_to_schema[field]
                ), f"Mismatched request context field types for {field}"
            else:
                self.arg_to_schema[field] = other.arg_to_schema[field]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/id.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/common/id.proto',
  package='tecton_proto.common',
  syntax='proto2',
  serialized_options=b'\n\021com.tecton.commonP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1ctecton_proto/common/id.proto\x12\x13tecton_proto.common\"n\n\x02Id\x12\x32\n\x15most_significant_bits\x18\x01 \x01(\x06R\x13mostSignificantBits\x12\x34\n\x16least_significant_bits\x18\x02 \x01(\x06R\x14leastSignificantBitsB\x15\n\x11\x63om.tecton.commonP\x01'
)




_ID = _descriptor.Descriptor(
  name='Id',
  full_name='tecton_proto.common.Id',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='most_significant_bits', full_name='tecton_proto.common.Id.most_significant_bits', index=0,
      number=1, type=6, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='mostSignificantBits', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='least_significant_bits', full_name='tecton_proto.common.Id.least_significant_bits', index=1,
      number=2, type=6, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='leastSignificantBits', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=163,
)

DESCRIPTOR.message_types_by_name['Id'] = _ID
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Id = _reflection.GeneratedProtocolMessageType('Id', (_message.Message,), {
  'DESCRIPTOR' : _ID,
  '__module__' : 'tecton_proto.common.id_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.common.Id)
  })
_sym_db.RegisterMessage(Id)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import glob
import importlib
import io
import os
import sys
import tarfile
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Callable
from typing import List
from typing import Tuple
from typing import Union

import click
import pytest
import requests
from google.protobuf.empty_pb2 import Empty
from yaspin.spinners import Spinners

import tecton
from .cli_utils import bold
from .cli_utils import confirm_or_exit
from tecton._internals import metadata_service
from tecton._internals import sdk_decorators
from tecton._internals.analytics import TectonStateUpdateMetrics
from tecton._internals.display import Displayable
from tecton._internals.tecton_errors import TectonAPIInaccessibleError
from tecton._internals.utils import format_freshness_table
from tecton._internals.utils import format_materialization_attempts
from tecton._internals.utils import get_all_freshness
from tecton.cli import api_key
from tecton.cli import common
from tecton.cli import printer
from tecton.cli import snowflake
from tecton.cli import workspace
from tecton.cli.command import _cluster_url
from tecton.cli.command import TectonCommand
from tecton.cli.command import TectonGroup
from tecton.cli.engine import dump_local_state
from tecton.cli.engine import update_tecton_state
from tecton.cli.error_utils import pretty_error
from tecton.okta import OktaAuthorizationFlow
from tecton_proto.metadataservice.metadata_service_pb2 import GetFeatureViewRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetMaterializationStatusRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetRestoreInfoRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetStateUpdateLogRequest
from tecton_spark import repo_file_handler
from tecton_spark.fco_container import FcoContainer

CONTEXT_SETTINGS = {
    "max_content_width": 160,
    "help_option_names": ["-h", "--help"],
}


@click.group(name="tecton", context_settings=CONTEXT_SETTINGS, cls=TectonGroup)
@click.option("--verbose/--no-verbose", default=False, help="Be Verbose.")
@click.option("--debug/--no-debug", default=False, help="Enable debug info.")
@click.pass_context
def cli(ctx, verbose, debug):
    "Tecton command-line tool."
    from tecton_spark.logger import set_logging_level
    import logging

    set_logging_level(logging.ERROR)
    sdk_decorators.disable_sdk_public_method_decorator()

    # add cwd to path
    sys.path.append("")


@cli.command(requires_auth=False)
def version():
    """Print version."""
    tecton.version.summary()


class EngineCommand(TectonCommand):
    def __init__(
        self, *args, apply: bool, upgrade_all: bool = False, destroy: bool = False, has_plan_id: bool = False, **kwargs
    ):
        @click.pass_context
        def callback(ctx, skip_tests, safety_checks, json_out, plan_id=None):
            args = EngineArgs(
                skip_tests=skip_tests,
                no_safety_checks=(not safety_checks),
                json_out=json_out,
                debug=common.get_debug(ctx),
            )
            if plan_id:
                args.plan_id = plan_id
            return run_engine(args, apply=apply, upgrade_all=upgrade_all, destroy=destroy)

        params = [
            # TODO(Add help)
            click.Option(["--skip-tests/--no-skip-tests"], default=False),
            click.Option(
                ["--safety-checks/--no-safety-checks"], default=True, help="Disable interactive safety checks."
            ),
            click.Option(
                ["--json-out"],
                default="",
                help="[BETA][not stable] Output the tecton state update diff (as JSON) to the file path provided.",
            ),
        ]
        if has_plan_id:
            params.append(
                click.Option(["--plan-id"], default=None, type=str, help="Id of a previously computed plan to apply.")
            )
        super().__init__(*args, callback=callback, params=params, uses_workspace=True, **kwargs)


cli.add_command(api_key.api_key)
cli.add_command(snowflake.snowflake)
cli.add_command(workspace.workspace)
cli.add_command(
    EngineCommand(
        name="plan",
        apply=False,
        help="Compare your local feature definitions with remote state and *show* the plan to bring them in sync.",
    )
)
cli.add_command(
    EngineCommand(
        name="apply",
        apply=True,
        has_plan_id=True,
        help="Compare your local feature definitions with remote state and *apply* local changes to the remote.",
    )
)
cli.add_command(
    EngineCommand(
        name="upgrade",
        apply=True,
        upgrade_all=True,
        help="Upgrade remote feature definitions.",
        hidden=True,
    )
)
cli.add_command(
    EngineCommand(
        name="destroy",
        destroy=True,
        apply=True,
        help="Destroy all objects on the server side.",
    )
)


@cli.command(uses_workspace=True)
@click.pass_context
def test(ctx):
    """Run Tecton tests."""
    run_tests(debug=common.get_debug(ctx))


def py_path_to_module(path: Path, repo_root: Path) -> str:
    return str(path.relative_to(repo_root))[: -len(".py")].replace("./", "").replace("/", ".").replace("\\", ".")


def plural(x, singular, plural):
    if x == 1:
        return singular
    else:
        return plural


Fco = Union[
    tecton.Entity,
    tecton.data_sources.data_source.BaseDataSource,
    tecton.transformations.transformation.Transformation,
    tecton.feature_views.FeatureDefinition,
    tecton.FeatureService,
]


def import_module_with_pretty_errors(
    file_path: Path,
    module_path: str,
    py_files: List[Path],
    repo_root: Path,
    debug: bool,
    before_error: Callable[[], None],
) -> ModuleType:
    from pyspark.sql.utils import AnalysisException

    try:
        module = importlib.import_module(module_path)
        if Path(module.__file__) != file_path:
            before_error()
            relpath = file_path.relative_to(repo_root)
            printer.safe_print(
                f"Python module name {bold(module_path)} ({relpath}) conflicts with module {module_path} from {module.__file__}. Please use a different name.",
                file=sys.stderr,
            )
            sys.exit(1)

        return module
    except AnalysisException as e:
        before_error()
        pretty_error(
            Path(file_path),
            py_files,
            exception=e,
            repo_root=repo_root,
            error_message="Analysis error",
            error_details=e.desc,
            debug=debug,
        )
        sys.exit(1)
    except tecton.tecton_errors.TectonValidationError as e:
        before_error()
        pretty_error(Path(file_path), py_files, exception=e, repo_root=repo_root, error_message=e.args[0], debug=debug)
        sys.exit(1)
    except SyntaxError as e:
        before_error()
        details = None
        if e.text and e.offset:
            details = e.text + (" " * e.offset) + "^^^"
        pretty_error(
            Path(file_path),
            py_files,
            exception=e,
            repo_root=repo_root,
            error_message=e.args[0],
            error_details=details,
            debug=debug,
        )
        sys.exit(1)
    except TectonAPIInaccessibleError as e:
        before_error()
        printer.safe_print("Failed to connect to Tecton server at", e.args[1], ":", e.args[0])
        sys.exit(1)
    except Exception as e:
        before_error()
        pretty_error(Path(file_path), py_files, exception=e, repo_root=repo_root, error_message=e.args[0], debug=debug)
        sys.exit(1)


def collect_top_level_objects(py_files: List[Path], repo_root: Path, debug: bool, pretty_errors: bool) -> List[Fco]:
    from tecton._internals.fco import _ALL_FCOS

    modules = [py_path_to_module(p, repo_root) for p in py_files]

    with printer.safe_yaspin(Spinners.earth, text="Importing feature repository modules") as sp:
        for file_path, module_path in zip(py_files, modules):
            sp.text = f"Processing feature repository module {module_path}"

            if pretty_errors:
                module = import_module_with_pretty_errors(
                    file_path=file_path,
                    module_path=module_path,
                    py_files=py_files,
                    repo_root=repo_root,
                    debug=debug,
                    before_error=lambda: sp.fail(printer.safe_string("⛔")),
                )
            else:
                module = importlib.import_module(module_path)

        num_modules = len(modules)
        sp.text = (
            f"Imported {num_modules} Python {plural(num_modules, 'module', 'modules')} from the feature repository"
        )
        sp.ok(printer.safe_string("✅"))

        return list(_ALL_FCOS.values())


def prepare_args(debug: bool) -> Tuple[List[Fco], str, List[Path]]:
    repo_file_handler.ensure_prepare_repo()
    repo_files = repo_file_handler.repo_files()
    repo_root = repo_file_handler.repo_root()

    py_files = [p for p in repo_files if p.suffix == ".py"]
    os.chdir(repo_root)

    top_level_objects = collect_top_level_objects(py_files, repo_root=Path(repo_root), debug=debug, pretty_errors=True)

    return top_level_objects, repo_root, repo_files


def check_version():
    try:
        metadata_service.instance().Nop(Empty())
    except Exception as e:
        printer.safe_print("Error connecting to tecton server: ", e, file=sys.stderr)
        sys.exit(1)


@cli.command(hidden=True)
@click.pass_context
def dump(ctx) -> None:
    """Print debug info."""
    top_level_objects, _, _ = prepare_args(common.get_debug(ctx))
    dump_local_state(top_level_objects)


def get_test_paths(repo_root) -> List[str]:
    # Be _very_ careful updating this:
    #    `glob.glob` does bash-style globbing (ignores hidden files)
    #    `pathlib.Path.glob` does _not_ do bash-style glob (it shows hidden)
    #
    # Ignoring hidden files is a very important expectation for our usage of
    # pytest. Otherwise, we may test files that user does not intend us to
    # (like in their .git or .tox directories).
    #
    # NOTE: This won't filter out hidden files for Windows. Potentially:
    #    `bool(os.stat(filepath).st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN)`
    # would filter hidden files for Windows, but this would need some testing.
    candidate_test_files = glob.iglob(f"{repo_root}/**/tests/**/*.py", recursive=True)

    VIRTUAL_ENV = os.getenv("VIRTUAL_ENV")
    if VIRTUAL_ENV:
        return list(filter(lambda f: not f.startswith(VIRTUAL_ENV), candidate_test_files))

    return list(candidate_test_files)


def run_tests(debug: bool):
    repo_root = repo_file_handler._maybe_get_repo_root()
    if repo_root is None:
        printer.safe_print("Tecton tests must be run from a feature repo initialized using 'tecton init'!")
        sys.exit(1)

    prepare_args(debug)

    tests = get_test_paths(repo_root)
    if len(tests) == 0:
        printer.safe_print("⚠️  Running Tests: No tests found.")
        return

    os.chdir(repo_root)
    args = ["-s"] + tests

    printer.safe_print("🏃 Running Tests")
    exitcode = pytest.main(args)

    if exitcode == 5:
        # https://docs.pytest.org/en/stable/usage.html#possible-exit-codes
        printer.safe_print("⚠️  Running Tests: No tests found.")
        return None
    elif exitcode != 0:
        printer.safe_print("⛔ Running Tests: Tests failed :(")
        sys.exit(1)
    else:
        printer.safe_print("✅ Running Tests: Tests passed!")


# TODO: This class was created to match the old style arg parse struct when we were migrating to click to avoid having
# do a deep refactoring of the code that depends on it. It should be replaced.
@dataclass
class EngineArgs:
    skip_tests: bool
    no_safety_checks: bool
    json_out: str

    debug: bool


def run_engine(args: EngineArgs, apply: bool = False, destroy=False, upgrade_all=False) -> TectonStateUpdateMetrics:
    check_version()

    # Resolve the json_out_filename prior to running `prepare_args(...)` so
    # that relative directories in the file name are supported (`prepare_args`
    # changes the working directory).
    json_out_path = None
    if args.json_out:
        json_out_path = Path(args.json_out).resolve()

    # Must use hasattr instead of args.plan_id, because only `apply` has the plan_id arg, but this
    # code path is also used by `plan`, `destroy`, and `upgrade`, which will fail on args.plan_id
    plan_id = None
    if hasattr(args, "plan_id"):
        plan_id = args.plan_id

    if destroy:
        top_level_objects: List[Fco] = []
        repo_root = None
        repo_files: List[Path] = []
    else:
        top_level_objects, repo_root, repo_files = prepare_args(args.debug)

    if not (args.skip_tests or apply and plan_id):
        run_tests(args.debug)

    return update_tecton_state(
        objects=top_level_objects,
        apply=apply,
        debug=args.debug,
        interactive=not args.no_safety_checks,
        repo_files=repo_files,
        repo_root=repo_root,
        upgrade_all=upgrade_all,
        json_out_path=json_out_path,
        plan_id=plan_id,
    )


@cli.command()
def init() -> None:
    """Initialize feature repo."""
    init_feature_repo()


def init_feature_repo() -> None:
    if Path().resolve() == Path.home():
        printer.safe_print("You cannot set feature repository root to the home directory", file=sys.stderr)
        sys.exit(1)

    # If .tecton exists in a parent or child directory, error out.
    repo_root = repo_file_handler._maybe_get_repo_root()
    if repo_root not in [Path().resolve(), None]:
        printer.safe_print(".tecton already exists in a parent directory:", repo_root)
        sys.exit(1)

    child_dir_matches = list(Path().rglob("*/.tecton"))
    if len(child_dir_matches) > 0:
        dirs_str = "\n\t".join(map(lambda c: str(c.parent.resolve()), child_dir_matches))
        printer.safe_print(f".tecton already exists in child directories:\n\t{dirs_str}")
        sys.exit(1)

    dot_tecton = Path(".tecton")
    if not dot_tecton.exists():
        dot_tecton.touch()
        printer.safe_print("Local feature repository root set to", Path().resolve(), "\n", file=sys.stderr)
        printer.safe_print("💡 We recommend tracking this file in git:", Path(".tecton").resolve(), file=sys.stderr)
        printer.safe_print(
            "💡 Run `tecton apply` to apply the feature repository to the Tecton cluster.", file=sys.stderr
        )
    else:
        printer.safe_print("Feature repository is already set to", Path().resolve(), file=sys.stderr)


@cli.command(uses_workspace=True)
@click.argument("commit_id", required=False)
def restore(commit_id):
    """Restore feature repo state to that of past `tecton apply`.

    The commit to restore can either be passed as COMMIT_ID, or the latest will be used.
    """
    # Get the repo download URL from the metadata service.
    r = GetRestoreInfoRequest()
    r.workspace = common.get_current_workspace()
    if commit_id:
        r.commit_id = commit_id
    response = metadata_service.instance().GetRestoreInfo(r)

    # Download the repo.
    url = response.signed_url_for_repo_download
    commit_id = response.commit_id
    sdk_version = response.sdk_version
    # TODO: always print this message once enough customers are on new sdk versions
    sdk_version_msg = f"applied by SDK version {sdk_version}" if sdk_version else ""
    printer.safe_print(f"Restoring from commit {commit_id} {sdk_version_msg}")
    try:
        tar_response = requests.get(url)
        tar_response.raise_for_status()
    except requests.RequestException as e:
        raise SystemExit(e)

    # Find the repo root or initialize a default repot if not in a repo.
    root = repo_file_handler._maybe_get_repo_root()
    if not root:
        init_feature_repo()
        root = Path().resolve()
    repo_file_handler.ensure_prepare_repo()

    # Get user confirmation.
    repo_files = repo_file_handler.repo_files()
    if len(repo_files) > 0:
        for f in repo_files:
            printer.safe_print(f)
        confirm_or_exit("This operation may delete or modify the above files. Ok?")
        for f in repo_files:
            os.remove(f)

    # Extract the feature repo.
    with tarfile.open(fileobj=io.BytesIO(tar_response.content), mode="r|gz") as tar:
        for entry in tar:
            if os.path.isabs(entry.name) or ".." in entry.name:
                raise ValueError("Illegal tar archive entry")
            elif os.path.exists(root / Path(entry.name)):
                raise ValueError(f"tecton restore would overwrite an unexpected file: {entry.name}")
            tar.extract(entry, path=root)
    printer.safe_print("Success")


@cli.command(uses_workspace=True)
@click.option("--limit", default=10, type=int, help="Number of log entries to return.")
def log(limit):
    """View log of past `tecton apply`."""
    logRequest = GetStateUpdateLogRequest()
    logRequest.workspace = common.get_current_workspace()
    logRequest.limit = limit
    response = metadata_service.instance().GetStateUpdateLog(logRequest)
    for entry in response.entries:
        printer.safe_print(f"Commit:\t{entry.commit_id}")
        printer.safe_print(f"Author:\t{entry.applied_by}")
        printer.safe_print(f"Date:\t{entry.applied_at.ToDatetime()}")
        printer.safe_print()


@cli.command(requires_auth=False)
@click.argument("tecton_url", required=False)
@click.option(
    "--manual/--no-manual",
    default=False,
    help="Manually require user to open browser and paste login token. Needed when using the Tecton CLI in a headless environment.",
)
def login(tecton_url, manual):
    """Log in and authenticate Tecton CLI.

    The Tecton URL may be optionally passed on the command line as TECTON_URL, otherwise you will be prompted."""
    from urllib.parse import urlparse, urljoin
    from tecton import conf

    host = _cluster_url()

    if tecton_url is None:
        printer.safe_print("Enter configuration. Press enter to use current value")
        prompt = "Tecton Cluster URL [%s]: " % (host or "no current value. example: https://yourco.tecton.ai")
        new_host = input(prompt).strip()
        if new_host:
            host = new_host
    else:
        host = tecton_url
    try:
        urlparse(host)
    except:
        printer.safe_print("Tecton Cluster URL must be a valid URL")
        sys.exit(1)
    # add this check for now since it can be hard to debug if you don't specify https and API_SERVICE fails
    if host is None or not (host.startswith("https://") or host.startswith("http://localhost:")):
        if host is not None and "//" not in host:
            host = f"https://{host}"
        else:
            printer.safe_print("Tecton Cluster URL must start with https://")
            sys.exit(1)

    # find the cli's client id
    okta_config_url = urljoin(host, "app/okta-config.json")

    try:
        response = requests.get(okta_config_url)
        response.raise_for_status()
    except requests.RequestException as e:
        raise SystemExit(e)
    cli_client_id = response.json()["OKTA_CLI_CLIENT_ID"]
    conf.set("CLI_CLIENT_ID", cli_client_id)

    flow = OktaAuthorizationFlow(hands_free=not manual)
    auth_code, code_verifier, redirect_uri = flow.get_authorization_code()
    access_token, _, refresh_token, access_token_expiration = flow.get_tokens(auth_code, code_verifier, redirect_uri)
    if not access_token:
        printer.safe_print("Unable to obtain Tecton credentials")
        sys.exit(1)

    conf.set("API_SERVICE", urljoin(host, "api"))
    # FEATURE_SERVICE and API_SERVICE are expected to have the same base URI: <host>/api
    conf.set("FEATURE_SERVICE", conf.get_or_none("API_SERVICE"))
    conf.set("CLI_CLIENT_ID", cli_client_id)
    conf.set("OAUTH_ACCESS_TOKEN", access_token)
    if refresh_token is not None:
        conf.set("OAUTH_REFRESH_TOKEN", refresh_token)
    conf.set("OAUTH_ACCESS_TOKEN_EXPIRATION", access_token_expiration)

    conf.save_tecton_configs()
    conf.save_okta_tokens(access_token, access_token_expiration, refresh_token)
    printer.safe_print(f"✅ Updated configuration at {conf._LOCAL_TECTON_CONFIG_FILE}")


@cli.command()
def freshness():
    """Show cluster-wide feature freshness states."""
    # TODO: use GetAllFeatureFreshnessRequest once we implement Chronosphere based API.
    freshness_statuses = get_all_freshness(common.get_current_workspace())
    num_fvs = len(freshness_statuses)
    if num_fvs == 0:
        printer.safe_print("No Feature Views found in this workspace.")
        return

    printer.safe_print(format_freshness_table(freshness_statuses))


@cli.command()
@click.pass_context
@click.argument("feature_view_name")
@click.option("--limit", default=100, type=int, help="Set the maximum limit of results.")
@click.option("--errors-only/--no-errors-only", default=False, help="Only show errors.")
def materialization_status(ctx, feature_view_name, limit, errors_only):
    """Show materialization status information for a FeatureView with FEATURE_VIEW_NAME in the 'prod' workspace.

    Prepend the --verbose flag for more information."""
    # Fetch FeatureView
    fvRequest = GetFeatureViewRequest()
    fvRequest.version_specifier = feature_view_name
    fvRequest.workspace = common.get_current_workspace()
    fvRequest.disable_legacy_response = True
    fvResponse = metadata_service.instance().GetFeatureView(fvRequest)
    fco_container = FcoContainer(fvResponse.fco_container)
    fv_proto = fco_container.get_single_root()
    if not fv_proto:
        printer.safe_print(f"Feature view '{feature_view_name}' not found.")
        sys.exit(1)
    fv_id = fv_proto.feature_view_id

    # Fetch Materialization Status
    statusRequest = GetMaterializationStatusRequest()
    statusRequest.feature_package_id.CopyFrom(fv_id)
    statusResponse = metadata_service.instance().GetMaterializationStatus(statusRequest)

    column_names, materialization_status_rows = format_materialization_attempts(
        statusResponse.materialization_status.materialization_attempts,
        verbose=common.get_verbose(ctx),
        limit=limit,
        errors_only=errors_only,
    )

    printer.safe_print("All the displayed times are in UTC time zone")

    # Setting `max_width=0` creates a table with an unlimited width.
    table = Displayable.from_items(headings=column_names, items=materialization_status_rows, max_width=0)
    # Align columns in the middle horizontally
    table._text_table.set_cols_align(["c" for _ in range(len(column_names))])
    printer.safe_print(table)


def main():
    try:
        cli()
    finally:
        metadata_service.close_instance()


if __name__ == "__main__":
    main()



from typing import Dict
from typing import List
from typing import Union

from tecton._internals import errors
from tecton._internals import utils
from tecton._internals.fco import Fco
from tecton.basic_info import prepare_basic_info
from tecton.entities.entity_args import prepare_args
from tecton_proto.args.entity_pb2 import EntityArgs
from tecton_proto.args.repo_metadata_pb2 import SourceInfo
from tecton_proto.common.id_pb2 import Id
from tecton_spark import errors
from tecton_spark.logger import get_logger

logger = get_logger("Entity")


class Entity(Fco):
    """
    Declare an Entity, used to organize and join features.

    An Entity is a class that represents an Entity that is being modeled in Tecton.
    Entities are used to index and organize features - a :class:`FeatureView`
    contains at least one Entity.

    Entities contain metadata about *join keys*, which represent the columns
    that are used to join features together.

    Example of an Entity declaration:

    .. code-block:: python

        from tecton import Entity

        customer = Entity(name='churned_customer', join_keys=['customer_id'],
                    description='A customer subscribing to a Sports TV subscription service',
                    family='sports_channels',
                    owner='jules@tecton.ai',
                    tags={'release': 'development'})
    """

    _args: EntityArgs
    _source_info: SourceInfo

    def __init__(
        self,
        *,
        name: str,
        description: str = "",
        family: str = "",
        tags: Dict[str, str] = None,
        owner: str = "",
        default_join_keys: Union[str, List[str]] = None,
        join_keys: Union[str, List[str]] = None,
    ):
        """
        Declare a new Entity.

        :param name: Unique name for the new entity.
        :param description: Short description of the new entity.
        :param family: (Optional) Family of this Entity, used to group Tecton Objects.
        :param tags: (Optional) Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
        :param owner: Owner name (typically the email of the primary maintainer).
        :param default_join_keys: Deprecated. Use ``join_keys``. Names of columns that uniquely identify the entity in FeatureView's SQL statement
            for which features should be aggregated.
        :param join_keys: Names of columns that uniquely identify the entity in FeatureView's SQL statement
            for which features should be aggregated.

        :raises TectonValidationError: if the input parameters are invalid.
        """
        from tecton.cli.common import get_fco_source_info

        if default_join_keys is None and join_keys is None:
            raise errors.TectonValidationError("You must specify `join_keys` to create an Entity object.")

        basic_info = prepare_basic_info(name=name, description=description, owner=owner, family=family, tags=tags)

        resolved_join_keys = join_keys if join_keys is not None else default_join_keys
        if isinstance(resolved_join_keys, str):
            resolved_join_keys = [resolved_join_keys]

        args = prepare_args(join_keys=resolved_join_keys, basic_info=basic_info)

        self._source_info = get_fco_source_info()
        self._args = args

        Fco._register(self)

    @property
    def name(self) -> str:
        """
        Name of the entity.
        """
        return self._args.info.name

    @property
    def join_keys(self) -> List[str]:
        """
        Join keys of the entity.
        """
        return list(self._args.join_keys)

    def with_join_keys(self, join_keys: Union[str, List[str]]) -> "OverriddenEntity":
        """
        Returns a new OverriddenEntity object with updated join keys. The current entity class is not modified.

        :param join_keys: Names of columns that uniquely identify the entity in FeatureView's transformation
            for which features should be aggregated. Existing `join_keys` will be fully ignored.

        :raises TectonValidationError: if the input parameters are invalid.
        """
        if not isinstance(join_keys, list):
            join_keys = [join_keys]

        utils.validate_join_keys(join_keys)
        if len(join_keys) != len(self.join_keys):
            raise errors.ENTITY_INVALID_JOIN_KEYS_OVERRIDE(self.name)

        return OverriddenEntity(self, join_keys)

    def _id(self) -> Id:
        return self._args.entity_id


class OverriddenEntity:
    """
    Represents a pre-registered Tecton Entity object with overriden fields (join_keys).

    `OverriddenEntity` instance is not persisted in the Database.
    """

    def __init__(self, entity: Entity, new_join_keys: List[str]):
        """Instantiates a new OverriddenEntity object."""
        self._entity = entity
        self._overridden_join_keys = new_join_keys

    @property
    def name(self) -> str:
        """
        Name of the entity.
        """
        return self._entity.name

    @property
    def join_keys(self):
        """
        Overriden join keys of the entity.
        """
        return list(self._overridden_join_keys)

    def _id(self) -> Id:
        return self._entity._id()



from datetime import datetime
from io import BytesIO
from typing import Mapping
from typing import Optional
from typing import Union

import numpy as np
import pandas
import pendulum
import pyspark
import requests

from tecton import conf
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.sdk_decorators import sdk_public_method
from tecton._internals.utils import validate_join_key_types
from tecton.feature_services.query_helper import _QueryHelper
from tecton.interactive.data_frame import DataFrame
from tecton.interactive.data_frame import FeatureVector
from tecton.interactive.feature_definition import FeatureDefinition
from tecton_proto.data import feature_view_pb2
from tecton_proto.metadataservice.metadata_service_pb2 import GetFeatureViewRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetNewIngestDataframeInfoRequest
from tecton_proto.metadataservice.metadata_service_pb2 import IngestDataframeRequest
from tecton_spark.fco_container import FcoContainer
from tecton_spark.logger import get_logger

logger = get_logger("FeatureTable")

__all__ = ["FeatureTable", "get_feature_table"]


class FeatureTable(FeatureDefinition):
    """
    FeatureTable class.

    To get a FeatureTable instance, call :py:func:`tecton.get_feature_table`.
    """

    _proto: feature_view_pb2.FeatureView

    def __init__(self, proto, fco_container):
        """
        :param proto: FT proto
        :param fco_container: Contains all FV dependencies, e.g., Entities, DS-es, Transformations
        """
        self._proto = proto
        assert isinstance(fco_container, FcoContainer), type(fco_container)
        self._fco_container = fco_container

    @classmethod
    def _fco_type_name_singular_snake_case(cls) -> str:
        return "feature_table"

    @classmethod
    def _fco_type_name_plural_snake_case(cls) -> str:
        return "feature_tables"

    def __str__(self):
        return f"FeatureTable|{self.id}"

    def __repr__(self):
        return f"FeatureTable(name='{self.name}')"

    def timestamp_key(self) -> str:
        """
        Returns the timestamp_key column name of this FeatureTable.
        """
        return self._proto.timestamp_key

    @sdk_public_method
    def get_features(
        self,
        entities: Union[pyspark.sql.dataframe.DataFrame, pandas.DataFrame, None] = None,
        start_time: Optional[Union[pendulum.DateTime, datetime]] = None,
        end_time: Optional[Union[pendulum.DateTime, datetime]] = None,
    ) -> DataFrame:
        """
        Deprecated.
        Gets all features that are defined by this FeatureTable.

        :param entities: (Optional) A DataFrame that is used to filter down feature values.
            If specified, this DataFrame should only contain join key columns.
        :param start_time: (Optional) The interval start time from when we want to retrieve features.
        :param end_time:  (Optional) The interval end time until when we want to retrieve features.

        :return: A Tecton DataFrame with features values.
        """
        logger.warning(
            "Deprecated. Use the 'get_historical_features' method instead to view historical values for this feature table. "
            + "See the api reference: https://docs.tecton.ai/api-reference/stubs/tecton.interactive.FeatureTable.html#tecton.interactive.FeatureTable.get_historical_features"
        )

        return self.get_historical_features(start_time=start_time, end_time=end_time, entities=entities)

    @sdk_public_method
    def get_historical_features(
        self,
        spine: Optional[Union[pyspark.sql.dataframe.DataFrame, pandas.DataFrame, DataFrame]] = None,
        timestamp_key: Optional[str] = None,
        entities: Optional[Union[pyspark.sql.dataframe.DataFrame, pandas.DataFrame, DataFrame]] = None,
        start_time: Optional[Union[pendulum.DateTime, datetime]] = None,
        end_time: Optional[Union[pendulum.DateTime, datetime]] = None,
        save: bool = False,
        save_as: Optional[str] = None,
    ) -> DataFrame:
        """
        Returns a Tecton :class:`DataFrame` of historical values for this feature table.
        If no arguments are passed in, all feature values for this feature table will be returned in a Tecton DataFrame.
        Note:
        The `timestamp_key` parameter is only applicable when a spine is passed in.
        Parameters `start_time`, `end_time`, and `entities` are only applicable when a spine is not passed in.

        :param spine: (Optional) The spine to join against, as a dataframe.
            If present, the returned DataFrame will contain rollups for all (join key, temporal key)
            combinations that are required to compute a full frame from the spine.
            To distinguish between spine columns and feature columns, feature columns are labeled as
            `feature_view_name.feature_name` in the returned DataFrame.
            If spine is not specified, it'll return a DataFrame of  feature values in the specified time range.
        :type spine: Union[pyspark.sql.DataFrame, pandas.DataFrame, tecton.DataFrame]
        :param timestamp_key: (Optional) Name of the time column in spine.
            This method will fetch the latest features computed before the specified timestamps in this column.
            If unspecified, will default to the time column of the spine if there is only one present.
        :type timestamp_key: str
        :param entities: (Optional) A DataFrame that is used to filter down feature values.
            If specified, this DataFrame should only contain join key columns.
        :type entities: Union[pyspark.sql.DataFrame, pandas.DataFrame, tecton.DataFrame]
        :param start_time: (Optional) The interval start time from when we want to retrieve features.
            If no timezone is specified, will default to using UTC.
        :type start_time: Union[pendulum.DateTime, datetime.datetime]
        :param end_time:  (Optional) The interval end time until when we want to retrieve features.
            If no timezone is specified, will default to using UTC.
        :type end_time: Union[pendulum.DateTime, datetime.datetime]
        :param save: (Optional) Whether to persist the DataFrame as a Dataset object. Default is False.
        :type save: bool
        :param save_as: (Optional) name to save the DataFrame as.
                If unspecified and save=True, a name will be generated.
        :type save_as: str

        Examples:
            A FeatureTable :py:mod:`ft` with join key :py:mod:`user_id`.

            1) :py:mod:`ft.get_historical_features(spine)` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'date': [datetime(...), datetime(...), datetime(...)]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the specified timestamps in the spine.

            2) :py:mod:`ft.get_historical_features(spine, save_as='my_dataset)` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'date': [datetime(...), datetime(...), datetime(...)]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the specified timestamps in the spine. Save the DataFrame as dataset with the name :py:mod`my_dataset`.

            3) :py:mod:`ft.get_historical_features(spine, timestamp_key='date_1')` where :py:mod:`spine=pandas.Dataframe({'user_id': [1,2,3], 'date_1': [datetime(...), datetime(...), datetime(...)], 'date_2': [datetime(...), datetime(...), datetime(...)]})`
            Fetch historical features from the offline store for users 1, 2, and 3 for the specified timestamps in the 'date_1' column in the spine.

            4) :py:mod:`ft.get_historical_features(start_time=datetime(...), end_time=datetime(...))`
            Fetch all historical features from the offline store in the time range specified by `start_time` and `end_time`.

        :return: A Tecton DataFrame with features values.
        """

        return self._get_historical_features(
            spine,
            timestamp_key,
            start_time,
            end_time,
            entities,
            from_source=False,
            save=save,
            save_as=save_as,
        )

    @sdk_public_method
    def get_feature_vector(
        self,
        join_keys: Optional[Mapping[str, Union[int, np.int_, str, bytes]]] = None,
        include_join_keys_in_response: bool = False,
    ) -> FeatureVector:
        """
        Deprecated.
        Returns a single Tecton FeatureVector from the Online Store.

        :param join_keys: Join keys of the enclosed FeatureTable.
        :param include_join_keys_in_response: Whether to include join keys as part of the response FeatureVector.

        :return: A FeatureVector of the results.
        """
        logger.warning(
            "Deprecated. Use the 'get_online_features' method instead to fetch features from the Online Store for this feature table. "
            + "See the api reference: https://docs.tecton.ai/api-reference/stubs/tecton.interactive.FeatureTable.html#tecton.interactive.FeatureTable.get_feature_vector"
        )
        # doing checks here instead of get_online_features in order to provide the correct error messages
        if not self._proto.feature_table.online_enabled:
            raise errors.UNSUPPORTED_OPERATION(
                "get_feature_vector", "online_serving_enabled was not defined for this Feature View."
            )
        if join_keys is None:
            raise errors.FS_GET_FEATURE_VECTOR_REQUIRED_ARGS
        return self.get_online_features(join_keys, include_join_keys_in_response)

    @sdk_public_method
    def get_online_features(
        self,
        join_keys: Mapping[str, Union[int, np.int_, str, bytes]],
        include_join_keys_in_response: bool = False,
    ) -> FeatureVector:
        """
        Returns a single Tecton FeatureVector from the Online Store.

        :param join_keys: Join keys of the enclosed FeatureTable.
        :param include_join_keys_in_response: Whether to include join keys as part of the response FeatureVector.

        Examples:
            A FeatureTable :py:mod:`ft` with join key :py:mod:`user_id`.

            1) :py:mod:`ft.get_online_features(join_keys={'user_id': 1})`
            Fetch the latest features from the online store for user 1.

            2) :py:mod:`ft.get_online_features(join_keys={'user_id': 1}, include_join_keys_in_respone=True)`
            Fetch the latest features from the online store for user 1 and include the join key information (user_id=1) in the returned FeatureVector.

        :return: A FeatureVector of the results.
        """
        if not self._proto.feature_table.online_enabled:
            raise errors.UNSUPPORTED_OPERATION(
                "get_online_features", "online_serving_enabled was not defined for this Feature Table."
            )
        validate_join_key_types(join_keys)

        return _QueryHelper(self._proto.fco_metadata.workspace, feature_view_name=self.name).get_feature_vector(
            join_keys, include_join_keys_in_response, request_context_map=None, request_context_schema=None
        )

    @sdk_public_method
    def ingest(self, df: Union[pyspark.sql.dataframe.DataFrame, pandas.DataFrame]):
        """
        Ingests a Dataframe into the FeatureTable. This method kicks off a materialization
        job to write the data into the offline and online store, depending on the Feature Table configuration.

        :param df: The Dataframe to be ingested. Has to conform to the FeatureTable schema.
        """
        get_upload_info_request = GetNewIngestDataframeInfoRequest()
        get_upload_info_request.feature_definition_id.CopyFrom(self._id_proto)
        upload_info_response = metadata_service.instance().GetNewIngestDataframeInfo(get_upload_info_request)
        df_path = upload_info_response.df_path
        upload_url = upload_info_response.signed_url_for_df_upload

        # We write in the native format and avoid converting Pandas <-> Spark due to partially incompatible
        # type system, in specifically missing Int in Pandas
        if isinstance(df, pyspark.sql.dataframe.DataFrame):
            df.write.parquet(df_path)
        else:
            self._check_types_and_upload_df_pandas(upload_url, df_path, df)

        ingest_request = IngestDataframeRequest()
        ingest_request.workspace = self.workspace
        ingest_request.feature_definition_id.CopyFrom(self._id_proto)
        ingest_request.df_path = df_path
        response = metadata_service.instance().IngestDataframe(ingest_request)

    @sdk_public_method
    def materialization_status(self, verbose=False, limit=1000, sort_columns=None, errors_only=False):
        """
        Displays materialization information for the FeatureTable, which may include past jobs, scheduled jobs,
        and job failures.

        :param verbose: If set to true, method will display additional low level materialization information,
            useful for debugging.
        :param limit: Maximum number of jobs to return.
        :param sort_columns: A comma-separated list of column names by which to sort the rows.
        :param: errors_only: If set to true, method will only return jobs that failed with an error.
        """
        return self._materialization_status(verbose, limit, sort_columns, errors_only)

    def _upload_df_pandas(self, upload_url: str, df: pandas.DataFrame):
        out_buffer = BytesIO()
        df.to_parquet(out_buffer, index=False)

        # Maximum 1GB per ingestion
        if out_buffer.__sizeof__() > 1000000000:
            raise errors.FT_DF_TOO_LARGE

        r = requests.put(upload_url, data=out_buffer.getvalue())
        if r.status_code != 200:
            raise errors.FT_UPLOAD_FAILED(r.reason)

    def delete_keys(
        self,
        keys: Union[pyspark.sql.dataframe.DataFrame, pandas.DataFrame],
        online: bool = True,
        offline: bool = True,
    ) -> None:
        """
        Deletes any materialized data that matches the specified join keys from the FeatureTable.
        This method kicks off a job to delete the data in the offline and online stores.
        If a FeatureTable has multiple entities, the full set of join keys must be specified.
        Only supports Dynamo online store.
        Maximum 10000 keys can be deleted per request.

        :param keys: The Dataframe to be deleted. Must conform to the FeatureTable join keys.
        :param online: (Optional, default=True) Whether or not to delete from the online store.
        :param offline: (Optional, default=True) Whether or not to delete from the offline store.
        :return: None if deletion job was created successfully.
        """
        return self._delete_keys(keys, online, offline)

    def deletion_status(self, verbose=False, limit=1000, sort_columns=None, errors_only=False):
        """
        Displays information for deletion jobs created with the delete_keys() method,
        which may include past jobs, scheduled jobs, and job failures.

        :param verbose: If set to true, method will display additional low level deletion information,
            useful for debugging.
        :param limit: Maximum number of jobs to return.
        :param sort_columns: A comma-separated list of column names by which to sort the rows.
        :param: errors_only: If set to true, method will only return jobs that failed with an error.
        """
        return self._deletion_status(verbose, limit, sort_columns, errors_only)


@sdk_public_method
def get_feature_table(ft_reference: str, workspace_name: Optional[str] = None) -> FeatureTable:
    """
    Fetch an existing :class:`FeatureTable` by name.

    :param ft_reference: Either a name or a hexadecimal feature table ID.
    :returns: :class:`FeatureTable`
    """

    if workspace_name == None:
        logger.warning(
            "`tecton.get_feature_table('<name>')` is deprecated. Please use `tecton.get_workspace('<workspace_name>').get_feature_table('<name>')` instead."
        )

    request = GetFeatureViewRequest()
    request.version_specifier = ft_reference
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    request.disable_legacy_response = True
    response = metadata_service.instance().GetFeatureView(request)
    fco_container = FcoContainer(response.fco_container)
    ft_proto = fco_container.get_single_root()

    if not ft_proto:
        raise errors.FCO_NOT_FOUND(FeatureTable, ft_reference)

    if not ft_proto.HasField("feature_table"):
        raise errors.FCO_NOT_FOUND_WRONG_TYPE(FeatureTable, ft_reference, "get_feature_table")

    return FeatureTable(ft_proto, fco_container)
